from tests.test_parser import get_parsed_filter_data
from pyffmpeg.generator import CodeGenerator
from pprint import pprint


def get_generated_method(filter_name: str) -> str:
    filter_data = get_parsed_filter_data(filter_name)

    code_generator = CodeGenerator(filter_data)
    return code_generator.generate()


def test_split_code_streams():
    method_code = get_generated_method("split")
    assert "    def split(self" in method_code


def test_overlay_code_streams():
    method_code = get_generated_method("overlay")
    assert '    def overlay(self, overlay_stream: "Stream"' in method_code


def test_split_code_all_options():
    method_code = get_generated_method("split")
    assert (
        '    def split(self, outputs: int | None = None) -> "FilterMultiOutput":'
        in method_code
    )


def test_overlay_code_some_options():
    method_code = get_generated_method("overlay")
    assert (
        "    def overlay(self, overlay_stream: \"Stream\", x: str | None = None, y: str | None = None, eof_action: Literal['repeat', 'endall', 'pass'] | int | None = None, eval: Literal['init', 'frame'] | int | None = None, shortest: bool | None = None"
        in method_code
    )



from pyffmpeg._utils import (
    escape_filter_option,
    escape_filter_description,
    escape_text_content,
    convert_kwargs_to_cmd_line_args,
)
from pyffmpeg.generated_filters import GeneratedFiltersMixin
from pyffmpeg.errors import Error
from typing import Any, Sequence
import subprocess


class Node:
    """Abstract base class for all components in the FFmpeg processing graph."""


class ProcessableNode(Node):
    """Nodes that can be further processed with filters."""

    def __init__(self, num_output_streams: int = 1):
        self.output_streams: list[Stream] = [
            Stream(self) for i in range(num_output_streams)
        ]


class RunnableNode(Node):
    def __init__(self):
        self.global_options: list[str] = []

    def global_args(self, *args) -> "RunnableNode":
        """Adds global options"""
        self.global_options.extend(args)
        return self

    def overwrite_output(self) -> "RunnableNode":
        """Adds global overwrite option"""
        self.global_options.append("-y")
        return self

    def _compile_global_kwargs(self, options_dict: dict) -> list[str]:
        """Converts kwargs to list"""
        args = []
        key_map = {
            "overwrite_output": "y",
            "log_level": "loglevel",
            "quiet": "loglevel",
        }
        for key, value in options_dict.items():
            if key == "quiet" and value is True:
                args.extend(["-loglevel", "quiet"])
                continue

            flag_name = key_map.get(key, key)

            if value is True:
                # -y with overwrite_output=True
                args.append(f"-{flag_name}")
            elif value is not None and value is not False:
                # handles for example log_level="error"
                args.extend([f"-{flag_name}", str(value)])
        return args

    def get_args(self, overwrite_output=False, **global_options) -> list[str]:
        """Builds command arguments"""
        global_options["overwrite_output"] = overwrite_output
        kwargs_args = self._compile_global_kwargs(global_options)
        sorter = GraphSorter(self)
        command_builder = CommandBuilder(
            sorter.sort(), self.global_options + kwargs_args
        )
        return command_builder.build_args()

    def compile(
        self, cmd: str = "ffmpeg", overwrite_output: bool = False, **global_options
    ) -> list[str]:
        """Builds command line arguments for invoking ffmpeg"""
        if isinstance(cmd, str):
            cmd = [cmd]
        elif not isinstance(cmd, list):
            cmd = list(cmd)
        return cmd + self.get_args(overwrite_output=overwrite_output, **global_options)

    def run(
        self,
        cmd: str | list[str] = "ffmpeg",
        capture_stdout: bool = False,
        capture_stderr: bool = False,
        input: bytes | None = None,
        quiet: bool = False,
        overwrite_output: bool = False,
        cwd: str | None = None,
        compile_function=None,
    ) -> tuple[bytes | None, bytes | None]:
        """Execute the ffmpeg command represented by a RunnableNode"""
        if not isinstance(self, RunnableNode):
            raise TypeError(f"Expected RunnableNode, got {type(self)}")

        compile_function = compile_function or self.__class__.compile
        cmdline = compile_function(
            self,
            cmd=cmd,
            overwrite_output=overwrite_output,
            quiet=quiet,
        )

        stdout = subprocess.PIPE if capture_stdout else None
        stderr = subprocess.PIPE if capture_stderr else None

        try:
            process = subprocess.run(
                cmdline,
                input=input,
                stdout=stdout,
                stderr=stderr,
                cwd=cwd,
                check=True,
            )
            return process.stdout, process.stderr
        except subprocess.CalledProcessError as e:
            raise Error(
                "ffmpeg error (see stderr output for detail)",
                stdout=e.stdout,
                stderr=e.stderr,
            )

    def run_async(
        self,
        cmd: str | list[str] = "ffmpeg",
        pipe_stdin: bool = False,
        pipe_stdout: bool = False,
        pipe_stderr: bool = False,
        input: bytes | None = None,
        quiet: bool = False,
        overwrite_output: bool = False,
        cwd: str | None = None,
    ) -> subprocess.Popen:
        """Runs ffmpeg process asynchronously"""
        args = self.compile(cmd, overwrite_output=overwrite_output)
        stdin_stream = subprocess.PIPE if pipe_stdin else None
        stdout_stream = subprocess.PIPE if pipe_stdout else None
        stderr_stream = subprocess.PIPE if pipe_stderr else None
        if quiet:
            stderr_stream = subprocess.STDOUT
            stdout_stream = subprocess.DEVNULL
        return subprocess.Popen(
            args,
            stdin=stdin_stream,
            stdout=stdout_stream,
            stderr=stderr_stream,
            cwd=cwd,
        )


class InputNode(ProcessableNode):
    """Nodes representing input files."""

    def __init__(self, filename: str, options: dict[str, Any] = None):
        super().__init__()
        self.filename: str = filename
        self.options: dict[str, Any] = options

    def __eq__(self, other: object) -> bool:
        if not isinstance(other, InputNode):
            return NotImplemented
        return self.filename == other.filename

    def __hash__(self) -> int:
        return hash(id(self))

    def get_input_args(self) -> list[str]:
        """Returns command args for this input"""
        options = self.options.copy()
        args = []

        if format := options.pop("format", None):
            args.extend(["-f", str(format)])
        if video_size := options.pop("video_size", None):
            if isinstance(video_size, (tuple, list)) and len(video_size) == 2:
                video_size = f"{video_size[0]}x{video_size[1]}"
            args.extend(["-video_size", str(video_size)])

        args.extend(convert_kwargs_to_cmd_line_args(options))
        args.extend(["-i", self.filename])

        return args


class OutputNode(RunnableNode):
    """Nodes representing output files."""

    def __init__(
        self,
        filename: str,
        inputs: list["Stream"],
        output_options: dict[str, str | list[str]] = {},
    ):
        super().__init__()
        self.inputs: list[Stream] = inputs
        self.filename: str = filename
        self.output_options: dict[str, str | list[str]] = output_options

    def __eq__(self, other: object) -> bool:
        if not isinstance(other, OutputNode):
            return NotImplemented
        return self.filename == other.filename and self.inputs == other.inputs

    def __hash__(self) -> int:
        return hash(id(self))

    def _normalize_output_options(self):
        """Replaces keys names in output_options from human readable (passed by the user)"""
        """to names existing in ffmpeg docs, prepares for later use to build command"""
        """Casts option values to string"""
        keys_names_mapping = {
            "video_bitrate": "b:v",
            "audio_bitrate": "b:a",
            "format": "f",
        }
        self.output_options = {
            keys_names_mapping.get(k, k): v for k, v in self.output_options.items()
        }

        video_size = self.output_options.get("video_size")
        if isinstance(video_size, Sequence) and not isinstance(video_size, str):
            try:
                width, height = video_size
            except ValueError:
                raise ValueError(
                    "video_size must contain exactly two elements: (width, height)"
                )
            self.output_options["video_size"] = f"{width}x{height}"

    def get_output_args(self, enforce_output_mapping) -> list[str]:
        """Builds command args representing the output"""
        """Generates args for output options"""
        """Generates args for mapping streams to the output if neccessary"""
        args: list[str] = []
        self._normalize_output_options()
        options = self.output_options.copy()
        format = options.pop("f", None)
        args.extend(convert_kwargs_to_cmd_line_args(options, sort=False))

        if (
            len(self.inputs) == 1
            and isinstance(self.inputs[0].source_node, InputNode)
            and not isinstance(self.inputs[0], (TypedStream, IndexedStream))
            and not enforce_output_mapping
        ):
            args.append(self.filename)
            return args

        for input in self.inputs:
            args.append("-map")
            args.append(
                f"[{input.index}]"
                if isinstance(input.source_node, FilterNode)
                else input.index
            )

        if format:
            args.extend(["-f", str(format)])

        args.append(self.filename)

        return args


class MergedOutputNode(RunnableNode):
    """Node representing multiple outputs"""

    def __init__(self, outputs: Sequence[OutputNode]):
        super().__init__()
        self.outputs: tuple[OutputNode] = tuple(outputs)


class SinkNode(RunnableNode):
    """
    Represents a graph terminal node that is a sink filter (e.g., nullsink, buffersink),
    not a file output.
    """

    def __init__(self, filter_node: "FilterNode"):
        super().__init__()
        self.filter_node = filter_node
        self.inputs = filter_node.inputs

    def __eq__(self, other: object) -> bool:
        if not isinstance(other, SinkNode):
            return NotImplemented
        return self.filter_node == other.filter_node

    def __hash__(self) -> int:
        return hash(self.filter_node)


class FilterNode(ProcessableNode):
    """Nodes representing filter operations."""

    def __init__(
        self,
        filter_name: str,
        positional_arguments: tuple[str],
        named_arguments: dict[str, Any],
        inputs: list["Stream"],
        num_output_streams: int = 1,
    ):
        super().__init__(num_output_streams)
        self.filter_name: str = filter_name
        self.positional_arguments: tuple = positional_arguments
        self.named_arguments: dict = named_arguments
        self.inputs: list[Stream] = inputs

    def __eq__(self, other: object) -> bool:
        if not isinstance(other, FilterNode):
            return NotImplemented
        return (
            self.filter_name == other.filter_name
            and self.positional_arguments == other.positional_arguments
            and self.named_arguments == other.named_arguments
            and self.inputs == other.inputs
        )

    def __hash__(self) -> int:
        return hash(id(self))

    def get_command_string(self) -> str:
        """Builds a command string based on this filter"""
        input_streams = [f"[{input.index}]" for input in self.inputs]
        output_streams = [f"[{output.index}]" for output in self.output_streams]

        postional_arguments = (str(arg) for arg in self.positional_arguments)

        named_arguments = []
        for name, value in sorted(self.named_arguments.items()):
            if value is None:
                continue

            if value is True:
                value = "true"
            elif value is False:
                value = "false"

            val_escaped = escape_filter_description(escape_filter_option(value))
            named_arguments.append(f"{name}={val_escaped}")

        all_arguments = [*postional_arguments, *named_arguments]
        arguments_string = ":".join(all_arguments)

        return f"{''.join(input_streams)}{self.filter_name}{f'=' if arguments_string else ''}{arguments_string}{''.join(output_streams)}"


class Stream(GeneratedFiltersMixin):
    """Represents a single output stream from a node."""

    def __init__(self, source_node: Node):
        self.source_node: Node = source_node
        self.index: str | None = None
        self.elementary_streams: dict[str, TypedStream | IndexedStream] = {}

    def __eq__(self, other: object) -> bool:
        if not isinstance(other, Stream):
            return NotImplemented
        return self.source_node == other.source_node

    def __hash__(self) -> int:
        return hash(id(self))

    @property
    def audio(self) -> "TypedStream":
        return self.elementary_streams.setdefault("a", TypedStream("audio", self))

    @property
    def video(self) -> "TypedStream":
        return self.elementary_streams.setdefault("v", TypedStream("video", self))

    def __getitem__(self, key: str) -> "TypedStream | IndexedStream":
        """Allows accessing elementary stream contained in this Stream"""
        mapping = {"a": "audio", "v": "video"}
        if not isinstance(key, str) or len(key) != 1:
            raise TypeError("Expected string index (e.g. 'a')")

        if key in mapping:
            stream_type = mapping[key]
            return self.elementary_streams.setdefault(
                key, TypedStream(stream_type, self)
            )

        if key.isdigit():
            return self.elementary_streams.setdefault(key, IndexedStream(self))

        raise KeyError(
            f"Invalid stream key: {key!r}. Expected 'a', 'v', or a numeric index like '1'."
        )

    @property
    def node(self):
        """Returns the list of output streams produced by the filter node
        that generated this stream."""
        if not isinstance(self.source_node, FilterNode):
            raise AttributeError(".node is only available on filter outputs")
        return self.source_node.output_streams

    def _apply_filter(
        self,
        filter_name: str,
        postional_arguments: tuple = (),
        named_arguments: dict[str, Any] = {},
        inputs: list["Stream"] | None = None,
        num_output_streams: int = 1,
    ) -> list["Stream"]:
        """Creates a FilterNode and returns its output streams."""
        node = FilterNode(
            filter_name,
            postional_arguments,
            named_arguments,
            inputs or [self],
            num_output_streams,
        )
        return node.output_streams

    def _apply_dynamic_outputs_filter(
        self,
        filter_name: str,
        postional_arguments: tuple = (),
        named_arguments: dict[str, Any] = {},
        inputs: list["Stream"] | None = None,
    ) -> "FilterMultiOutput":
        """Creates a FilterNode and returns FilterMultiOutput to allow dynamic outputs."""
        filter_node = FilterNode(
            filter_name,
            postional_arguments,
            named_arguments,
            inputs or [self],
            num_output_streams=0,
        )
        return FilterMultiOutput(filter_node)

    def _apply_sink_filter(
        self,
        filter_name: str,
        named_arguments: dict = {},
        inputs: list["Stream"] | None = None,
    ) -> SinkNode:
        filter_node = FilterNode(
            filter_name,
            positional_arguments=(),
            named_arguments=named_arguments,
            inputs=inputs or [self],
            num_output_streams=0,
        )
        return SinkNode(filter_node)

    def filter(self, filter_name: str, *args, **kwargs) -> "Stream":
        """Custom filter with a single input and a single output"""
        return self._apply_filter(filter_name, args, kwargs)[0]

    def filter_multi_output(
        self, filter_name: str, *args, **kwargs
    ) -> "FilterMultiOutput":
        """Creates a custom filter allowing dynamic creation of output streams"""
        node = FilterNode(
            filter_name=filter_name,
            positional_arguments=args,
            named_arguments=kwargs,
            inputs=[self],
            num_output_streams=0,
        )
        return FilterMultiOutput(node)

    def output(
        self,
        filename: str,
        streams: list["Stream"] | None = None,
        format: str | None = None,
        vcodec: str | None = None,
        acodec: str | None = None,
        video_bitrate: str | int | None = None,
        audio_bitrate: str | int | None = None,
        aspect: str | float | None = None,
        frames: int | None = None,
        shortest: bool = False,
        **kwargs,
    ) -> "OutputNode":
        """
        Creates an output node for this stream, optionally muxing other streams.

        Args:
            filename (str): The output file path or URL.
            streams (list[Stream] | None): Additional streams to include in the output
                (e.g., audio tracks, subtitles) alongside the current stream.
            format (str | None): Force output format (ffmpeg flag: ``-f``).
            vcodec (str | None): Video codec (ffmpeg flag: ``-c:v``).
            acodec (str | None): Audio codec (ffmpeg flag: ``-c:a``).
            video_bitrate (str | int | None): Video bitrate (ffmpeg flag: ``-b:v``).
            audio_bitrate (str | int | None): Audio bitrate (ffmpeg flag: ``-b:a``).
            aspect (str | float | None): Set aspect ratio (ffmpeg flag: ``-aspect``).
            frames (int | None): Number of video frames to output (ffmpeg flag: ``-frames:v``).
            shortest (bool): Finish encoding when the shortest input stream ends.
            **kwargs: Additional output options.

        Returns:
            OutputNode: The created output node.
        """

        output_streams = [self]

        if streams is not None:
            output_streams.extend(streams)

        options = kwargs.copy()

        if format is not None:
            options["f"] = format
        if vcodec is not None:
            options["c:v"] = vcodec
        if acodec is not None:
            options["c:a"] = acodec
        if video_bitrate is not None:
            options["b:v"] = video_bitrate
        if audio_bitrate is not None:
            options["b:a"] = audio_bitrate
        if aspect is not None:
            options["aspect"] = aspect
        if frames is not None:
            options["frames:v"] = frames
        if shortest:
            options["shortest"] = None

        return OutputNode(filename, output_streams, output_options=options)


class TypedStream(Stream):
    """Elementary stream representing specific type of media out of those contained within a Stream"""

    def __init__(self, type: str, source_stream: Stream):
        self.type = type
        self.source_node = source_stream.source_node
        self.index: str | None = None

    def __getitem__(self, _):
        """Raises ValueError"""
        mapping = {"audio": "a", "video": "v"}
        raise ValueError(f"Stream already has a selector: {mapping[self.type]}")


class IndexedStream(Stream):
    """Elementary stream represented by index within the containing stream"""

    def __init__(self, source_stream: Stream):
        self.source_node = source_stream.source_node
        self.label: str
        self.index: str | None = None

    def __getitem__(self, _):
        """Raises ValueError"""
        raise ValueError(f"Stream already has a selector")


class GraphSorter:
    def __init__(self, output: RunnableNode):
        self.start_node: RunnableNode = output
        self.visited: set[Node] = set()
        self.sorted: list[Node] = []
        self.current_input_stream_index = 0
        self.current_filter_stream_index = 0

    def sort(self) -> list[Node]:
        self._sort(self.start_node)
        type_order = {InputNode: 0, FilterNode: 1, SinkNode: 2, OutputNode: 3}
        self.sorted.sort(key=lambda x: type_order[type(x)])
        self.label_streams()
        return self.sorted

    def _sort(self, node: Node) -> None:
        if node in self.visited:
            return
        self.visited.add(node)

        if isinstance(node, (FilterNode, SinkNode, OutputNode)):
            for stream in node.inputs:
                self._sort(stream.source_node)

        if isinstance(node, MergedOutputNode):
            for output in node.outputs:
                self._sort(output)
            return

        self.sorted.append(node)

    def label_streams(self) -> None:
        for node in self.sorted:
            if isinstance(node, InputNode):
                for stream in node.output_streams:
                    stream.index = str(self.current_input_stream_index)
                    self.label_elementary_streams(stream)
                    self.current_input_stream_index += 1
            if isinstance(node, FilterNode):
                for stream in node.output_streams:
                    stream.index = f"s{self.current_filter_stream_index}"
                    self.label_elementary_streams(stream)
                    self.current_filter_stream_index += 1

    def label_elementary_streams(self, stream: Stream):
        for type, elementary_stream in stream.elementary_streams.items():
            elementary_stream.index = f"{stream.index}:{type}"


class CommandBuilder:
    def __init__(self, nodes: list[Node], global_options: list = []):
        self.nodes: list[Node] = nodes
        self.global_options: list[str] = global_options

    def build_args(self) -> list[str]:
        args: list[str] = []
        filters = []
        outputs = []
        filter_complex: bool = False
        multi_input: bool = True if isinstance(self.nodes[1], InputNode) else False
        enforce_output_mapping: bool = False
        for node in self.nodes:
            if isinstance(node, InputNode):
                args.extend(node.get_input_args())
            if isinstance(node, FilterNode):
                if not filter_complex:
                    args.append("-filter_complex")
                    filter_complex = True

                filters.append(node.get_command_string())

            if isinstance(node, SinkNode):
                if not filter_complex:
                    args.append("-filter_complex")
                    filter_complex = True
                filters.append(node.filter_node.get_command_string())

            if isinstance(node, OutputNode):
                outputs.extend(node.get_output_args(enforce_output_mapping))
                if multi_input:
                    enforce_output_mapping = True

        if filters:
            args.append(";".join(filters))
        args.extend(outputs)

        args.extend(self.global_options)

        return args


class FilterMultiOutput:
    """Filter node wrapper which allows creating arbitrary outputs for the filter node dynamically"""

    def __init__(self, filter_node: FilterNode):
        self.filter_node = filter_node
        self.stream_cache: dict[str, Stream] = {}

    def __getitem__(self, key: str | int) -> "Stream":
        """Returns new or existing stream under label"""
        label = str(key)
        if label in self.stream_cache:
            return self.stream_cache[label]

        new_stream = Stream(self.filter_node)
        self.filter_node.output_streams.append(new_stream)
        self.stream_cache[label] = new_stream
        return new_stream



from typing import Any
import re


class Parser:
    """Parses text output from 'ffmpeg --help filter=...' command."""

    def __init__(self, text: str) -> None:
        self.splitted_input: list[str] = text.splitlines()
        self.total_lines: int = len(self.splitted_input)
        self.current_index = 0
        self.line: str | None = None
        self.advance()
        self.filter_data = {}

    def advance(self) -> None:
        """Moves forward to the next line and buffers it."""
        if self.current_index == self.total_lines:
            self.line = None
            return
        self.line = self.splitted_input[self.current_index]
        self.current_index += 1

    def parse(self):
        """Orchestrates the parsing process."""
        if filter_name := self.parse_filter_name():
            self.filter_data["filter_name"] = filter_name
        if filter_description := self.parse_description():
            self.filter_data["description"] = filter_description
        self.filter_data["inputs"] = self.parse_inputs()
        self.filter_data["outputs"] = self.parse_outputs()
        if options := self.parse_options_block():
            self.filter_data["options"] = options
        return self.filter_data

    def parse_filter_name(self) -> str | None:
        """Scans the text for the 'Filter <name>' header"""
        while self.line is not None:
            if self.line.startswith("Filter "):
                parts = self.line.split()
                filter_name = parts[1]

                self.advance()
                return filter_name

            self.advance()

        return None

    def parse_description(self) -> str | None:
        """Gets filter description and skips to the next section"""
        description = None

        if self.line is not None and not self._is_section_header():
            description = self.line.strip()
            self.advance()

        while self.line is not None:
            if self._is_section_header():
                break
            self.advance()

        return description

    def parse_inputs(self) -> list[dict[str, str]]:
        """Parses the Inputs section."""
        inputs = []
        found_dynamic_keyword = False

        if self.line is None or self.line.strip() != "Inputs:":
            self.filter_data["is_dynamic_inputs"] = False
            return inputs

        self.advance()

        while self.line is not None:
            if self._is_section_header():
                break

            line = self.line.strip()

            if "none (source filter)" in line:
                self.advance()
                continue

            if "dynamic" in line.lower():
                found_dynamic_keyword = True
                self.advance()
                continue

            if line.startswith("#"):
                if input_data := self._parse_stream_line():
                    inputs.append(input_data)

            self.advance()

        if len(inputs) > 0:
            self.filter_data["is_dynamic_inputs"] = False
        else:
            self.filter_data["is_dynamic_inputs"] = found_dynamic_keyword

        return inputs

    def parse_outputs(self) -> list[dict[str, str]]:
        """Parses the Outputs section."""
        outputs = []
        found_dynamic_keyword = False

        if self.line is None or self.line.strip() != "Outputs:":
            return outputs
        self.advance()

        while self.line is not None:
            if self._is_section_header():
                break

            line = self.line.strip()

            if "none" in line.lower():
                self.advance()
                continue

            if "dynamic" in line.lower():
                found_dynamic_keyword = True
                self.advance()
                continue

            line = self.line.strip()
            if line.startswith("#"):
                if output_data := self._parse_stream_line():
                    outputs.append(output_data)

            self.advance()

        if len(outputs) > 0:
            self.filter_data["is_dynamic_outputs"] = False
        else:
            self.filter_data["is_dynamic_outputs"] = found_dynamic_keyword

        return outputs

    def _parse_stream_line(self) -> dict[str, str] | None:
        """
        Extracts name and type from line like: '#0: main (video)'
        Returns: {'name': 'main', 'type': 'video'} or None
        """
        parts = self.line.split(":")
        if len(parts) < 2:
            return None

        content = parts[1].strip()

        if "(" in content and content.endswith(")"):
            name_part, type_part = content.split("(", 1)
            return {"name": f"{name_part.strip()}_stream", "type": type_part.strip(")")}

        return None

    def parse_options_block(self) -> list[dict[str, Any]]:
        """Parses one block of AVOptions"""
        options = []
        if self.line is None or not self.line.strip().endswith("AVOptions:"):
            return options

        self.advance()

        while self.line is not None:
            stripped = self.line.strip()

            if self._is_section_header():
                break
            if not stripped:
                self.advance()
                continue

            parts = stripped.split(maxsplit=3)
            if len(parts) < 2:
                self.advance()
                continue

            name = parts[0]
            second_col = parts[1]
            raw_description = parts[3] if len(parts) > 3 else None

            if second_col.startswith("<") and second_col.endswith(">"):
                option_type = second_col[1:-1]

                description, default = self._extract_meta(raw_description)

                options.append(
                    {
                        "name": name,
                        "type": option_type,
                        "description": description,
                        "default": default,
                        "choices": [],
                    }
                )

            else:
                if options:
                    options[-1]["choices"].append(
                        {
                            "name": name,
                            "value": second_col,
                            "description": raw_description,
                        }
                    )
                else:
                    pass

            self.advance()

        return options

    def _extract_meta(self, text: str) -> tuple[str, str | None]:
        """Extracts description and default value."""
        if not text or "(default " not in text:
            return text, None

        description_part, value_part = text.rsplit("(default ", 1)
        raw_value = value_part.rsplit(")", 1)[0]
        default_val = raw_value.strip('"').strip()

        return description_part.strip(), default_val

    def _is_section_header(self) -> bool:
        """Checks if the line starts a new section."""
        if self.line is None:
            return False
        stripped = self.line.strip()
        return (
            stripped in ["Inputs:", "Outputs:"]
            or stripped.endswith("AVOptions:")
            or stripped.startswith("Exiting with")
            or stripped.startswith("This filter has support")
        )



import pytest


@pytest.fixture(params=["NEW", "COMPAT"])
def ffmpeg_module(request):
    if request.param == "COMPAT":
        import pyffmpeg._compat as ffmpeg

        return ffmpeg
    elif request.param == "NEW":
        import pyffmpeg as ffmpeg

        return ffmpeg
    else:
        raise NotImplementedError(request.param)



from pyffmpeg import _utils
from pyffmpeg.io import input, merge_outputs, get_args, filter, filter_multi_output
from pyffmpeg._run import run, compile
from pyffmpeg.errors import Error
from pyffmpeg.node import Stream, Node
from pyffmpeg.probe import probe
from collections.abc import Callable
from typing import Any


def __getattr__(name: str) -> Callable[..., Any]:
    def wrapper(obj: Stream | Node | list[Stream], *args, **kwargs) -> Any:
        target_obj = obj
        inputs_list = []

        if isinstance(obj, (list, tuple)) and obj:
            target_obj = obj[0]
            inputs_list = obj[1:]

        if not isinstance(target_obj, (Stream, Node)):
            raise TypeError(
                f"Cannot call '{name}'. First argument is '{type(obj)}', "
                f"but must be Stream, Node or a list of Streams."
            )

        try:
            method = getattr(target_obj, name)
        except AttributeError:
            raise AttributeError(
                f"Module 'pyffmpeg' does not have a function '{name}' "
                f"and an object of type '{type(target_obj)}' does not have a method with that name."
            )

        return method(*inputs_list, *args, **kwargs)

    return wrapper



class Error(Exception):
    """Error raised when running an ffmpeg process"""

    def __init__(
        self,
        message: str,
        stdout: bytes | None = None,
        stderr: bytes | None = None,
    ) -> None:
        super().__init__(message)
        self.stdout: bytes | None = stdout
        self.stderr: bytes | None = stderr



from typing import Any, Iterable


def escape_chars(text: str, characters: str) -> str:
    """Escapes given characters in the text"""
    escape_set = set(characters)
    result = []
    for character in str(text):
        if character in escape_set:
            result.append(f"\\{character}")
        else:
            result.append(character)
    return "".join(result)


def escape_filter_option(text: str) -> str:
    """Escapes text on ffmpeg filter option level"""
    return escape_chars(text, ":='\\")


def escape_filter_description(text: str) -> str:
    """Escapes text on ffmpeg filter description level"""
    return escape_chars(text, "[],;'\\")


def escape_text_content(text: str) -> str:
    """Escapes text on ffmpeg text values level"""
    """Only needed for special filter options that are textual"""
    return escape_chars(text, "\\%'")


def convert_kwargs_to_cmd_line_args(
    kwargs: dict[str, Any], sort: bool = True
) -> list[str]:
    """
    Converts kwargs to list of args.
    """
    args = []
    options = sorted(kwargs.items()) if sort else kwargs.items()
    for key, value in options:
        if isinstance(value, Iterable) and not isinstance(value, (str, bytes)):
            for v in value:
                args.append(f"-{key}")
                if v is not None:
                    args.append(str(v))
        else:
            args.append(f"-{key}")
            if value is not None:
                args.append(str(value))
    return args


from pyffmpeg.node import Stream, FilterMultiOutput, FilterNode


def create_source(
    filter_name: str,
    positional_arguments: tuple = (),
    named_arguments: dict[str, Any] = {},
    num_outputs: int = 1,
    is_dynamic: bool = False,
) -> Stream | list[Stream] | FilterMultiOutput:
    """
    Internal factory for creating source filters.
    Handles single, multiple, and dynamic outputs.
    """
    node = FilterNode(
        filter_name=filter_name,
        positional_arguments=positional_arguments,
        named_arguments=named_arguments,
        inputs=[],
        num_output_streams=0 if is_dynamic else num_outputs,
    )

    if is_dynamic:
        return FilterMultiOutput(node)

    if num_outputs == 1:
        return node.output_streams[0]

    return node.output_streams



# --- AUTO-GENERATED FILE ---
from typing import TYPE_CHECKING, Literal
from pyffmpeg._utils import create_source

if TYPE_CHECKING:
    from pyffmpeg.node import Stream, FilterMultiOutput


def abuffer(
    time_base: str | None = None,
    sample_rate: int | None = None,
    sample_fmt: str | None = None,
    channel_layout: str | None = None,
    channels: int | None = None,
) -> "Stream":
    """Buffer audio frames, and make them accessible to the filterchain.
    Args:
        time_base (str): (from 0 to INT_MAX)

            Defaults to 0/1.
        sample_rate (int): (from 0 to INT_MAX)

            Defaults to 0.
        sample_fmt (str): No description available.

            Defaults to none.
        channel_layout (str): No description available.

        channels (int): (from 0 to INT_MAX)

            Defaults to 0.

    Returns:
        "Stream": The output stream.
    """
    return create_source(
        filter_name="abuffer",
        named_arguments={
            "time_base": time_base,
            "sample_rate": sample_rate,
            "sample_fmt": sample_fmt,
            "channel_layout": channel_layout,
            "channels": channels,
        },
    )


def aevalsrc(
    exprs: str | None = None,
    nb_samples: int | None = None,
    n: int | None = None,
    sample_rate: str | None = None,
    s: str | None = None,
    duration: str | None = None,
    d: str | None = None,
    channel_layout: str | None = None,
    c: str | None = None,
) -> "Stream":
    """Generate an audio signal generated by an expression.
    Args:
        exprs (str): set the '|'-separated list of channels expressions

        nb_samples (int): set the number of samples per requested frame (from 0 to INT_MAX)

            Defaults to 1024.
        n (int): set the number of samples per requested frame (from 0 to INT_MAX)

            Defaults to 1024.
        sample_rate (str): set the sample rate

            Defaults to 44100.
        s (str): set the sample rate

            Defaults to 44100.
        duration (str): set audio duration

            Defaults to -0.000001.
        d (str): set audio duration

            Defaults to -0.000001.
        channel_layout (str): set channel layout

        c (str): set channel layout


    Returns:
        "Stream": The output stream.
    """
    return create_source(
        filter_name="aevalsrc",
        named_arguments={
            "exprs": exprs,
            "nb_samples": nb_samples,
            "n": n,
            "sample_rate": sample_rate,
            "s": s,
            "duration": duration,
            "d": d,
            "channel_layout": channel_layout,
            "c": c,
        },
    )


def afdelaysrc(
    delay: float | None = None,
    d: float | None = None,
    sample_rate: int | None = None,
    r: int | None = None,
    nb_samples: int | None = None,
    n: int | None = None,
    taps: int | None = None,
    t: int | None = None,
    channel_layout: str | None = None,
    c: str | None = None,
) -> "Stream":
    """Generate a Fractional delay FIR coefficients.
    Args:
        delay (float): set fractional delay (from 0 to 32767)

            Defaults to 0.
        d (float): set fractional delay (from 0 to 32767)

            Defaults to 0.
        sample_rate (int): set sample rate (from 1 to INT_MAX)

            Defaults to 44100.
        r (int): set sample rate (from 1 to INT_MAX)

            Defaults to 44100.
        nb_samples (int): set the number of samples per requested frame (from 1 to INT_MAX)

            Defaults to 1024.
        n (int): set the number of samples per requested frame (from 1 to INT_MAX)

            Defaults to 1024.
        taps (int): set number of taps for delay filter (from 0 to 32768)

            Defaults to 0.
        t (int): set number of taps for delay filter (from 0 to 32768)

            Defaults to 0.
        channel_layout (str): set channel layout

            Defaults to stereo.
        c (str): set channel layout

            Defaults to stereo.

    Returns:
        "Stream": The output stream.
    """
    return create_source(
        filter_name="afdelaysrc",
        named_arguments={
            "delay": delay,
            "d": d,
            "sample_rate": sample_rate,
            "r": r,
            "nb_samples": nb_samples,
            "n": n,
            "taps": taps,
            "t": t,
            "channel_layout": channel_layout,
            "c": c,
        },
    )


def afireqsrc(
    preset: Literal[
        "custom",
        "flat",
        "acoustic",
        "bass",
        "beats",
        "classic",
        "clear",
        "deep",
        "dubstep",
        "electronic",
        "hardstyle",
        "hip-hop",
        "jazz",
        "metal",
        "movie",
        "pop",
        "r&b",
        "rock",
        "vocal",
    ]
    | int
    | None = None,
    p: Literal[
        "custom",
        "flat",
        "acoustic",
        "bass",
        "beats",
        "classic",
        "clear",
        "deep",
        "dubstep",
        "electronic",
        "hardstyle",
        "hip-hop",
        "jazz",
        "metal",
        "movie",
        "pop",
        "r&b",
        "rock",
        "vocal",
    ]
    | int
    | None = None,
    gains: str | None = None,
    g: str | None = None,
    bands: str | None = None,
    b: str | None = None,
    taps: int | None = None,
    t: int | None = None,
    sample_rate: int | None = None,
    r: int | None = None,
    nb_samples: int | None = None,
    n: int | None = None,
    interp: Literal["linear", "cubic"] | int | None = None,
    i: Literal["linear", "cubic"] | int | None = None,
    phase: Literal["linear", "min"] | int | None = None,
    h: Literal["linear", "min"] | int | None = None,
) -> "Stream":
    """Generate a FIR equalizer coefficients audio stream.
    Args:
        preset (int | str): set equalizer preset (from -1 to 17)

            Allowed values:
                * custom
                * flat
                * acoustic
                * bass
                * beats
                * classic
                * clear
                * deep
                * dubstep
                * electronic
                * hardstyle
                * hip-hop
                * jazz
                * metal
                * movie
                * pop
                * r&b
                * rock
                * vocal

            Defaults to flat.
        p (int | str): set equalizer preset (from -1 to 17)

            Allowed values:
                * custom
                * flat
                * acoustic
                * bass
                * beats
                * classic
                * clear
                * deep
                * dubstep
                * electronic
                * hardstyle
                * hip-hop
                * jazz
                * metal
                * movie
                * pop
                * r&b
                * rock
                * vocal

            Defaults to flat.
        gains (str): set gain values per band

            Defaults to 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0.
        g (str): set gain values per band

            Defaults to 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0.
        bands (str): set central frequency values per band

            Defaults to 25 40 63 100 160 250 400 630 1000 1600 2500 4000 6300 10000 16000 24000.
        b (str): set central frequency values per band

            Defaults to 25 40 63 100 160 250 400 630 1000 1600 2500 4000 6300 10000 16000 24000.
        taps (int): set number of taps (from 16 to 65535)

            Defaults to 4096.
        t (int): set number of taps (from 16 to 65535)

            Defaults to 4096.
        sample_rate (int): set sample rate (from 1 to INT_MAX)

            Defaults to 44100.
        r (int): set sample rate (from 1 to INT_MAX)

            Defaults to 44100.
        nb_samples (int): set the number of samples per requested frame (from 1 to INT_MAX)

            Defaults to 1024.
        n (int): set the number of samples per requested frame (from 1 to INT_MAX)

            Defaults to 1024.
        interp (int | str): set the interpolation (from 0 to 1)

            Allowed values:
                * linear
                * cubic

            Defaults to linear.
        i (int | str): set the interpolation (from 0 to 1)

            Allowed values:
                * linear
                * cubic

            Defaults to linear.
        phase (int | str): set the phase (from 0 to 1)

            Allowed values:
                * linear
                * min

            Defaults to min.
        h (int | str): set the phase (from 0 to 1)

            Allowed values:
                * linear
                * min

            Defaults to min.

    Returns:
        "Stream": The output stream.
    """
    return create_source(
        filter_name="afireqsrc",
        named_arguments={
            "preset": preset,
            "p": p,
            "gains": gains,
            "g": g,
            "bands": bands,
            "b": b,
            "taps": taps,
            "t": t,
            "sample_rate": sample_rate,
            "r": r,
            "nb_samples": nb_samples,
            "n": n,
            "interp": interp,
            "i": i,
            "phase": phase,
            "h": h,
        },
    )


def afirsrc(
    taps: int | None = None,
    t: int | None = None,
    frequency: str | None = None,
    f: str | None = None,
    magnitude: str | None = None,
    m: str | None = None,
    phase: str | None = None,
    p: str | None = None,
    sample_rate: int | None = None,
    r: int | None = None,
    nb_samples: int | None = None,
    n: int | None = None,
    win_func: Literal[
        "rect",
        "bartlett",
        "hann",
        "hanning",
        "hamming",
        "blackman",
        "welch",
        "flattop",
        "bharris",
        "bnuttall",
        "bhann",
        "sine",
        "nuttall",
        "lanczos",
        "gauss",
        "tukey",
        "dolph",
        "cauchy",
        "parzen",
        "poisson",
        "bohman",
        "kaiser",
        "rect",
        "bartlett",
        "hann",
        "hanning",
        "hamming",
        "blackman",
        "welch",
        "flattop",
        "bharris",
        "bnuttall",
        "bhann",
        "sine",
        "nuttall",
        "lanczos",
        "gauss",
        "tukey",
        "dolph",
        "cauchy",
        "parzen",
        "poisson",
        "bohman",
        "kaiser",
    ]
    | int
    | None = None,
    w: Literal[
        "rect",
        "bartlett",
        "hann",
        "hanning",
        "hamming",
        "blackman",
        "welch",
        "flattop",
        "bharris",
        "bnuttall",
        "bhann",
        "sine",
        "nuttall",
        "lanczos",
        "gauss",
        "tukey",
        "dolph",
        "cauchy",
        "parzen",
        "poisson",
        "bohman",
        "kaiser",
        "rect",
        "bartlett",
        "hann",
        "hanning",
        "hamming",
        "blackman",
        "welch",
        "flattop",
        "bharris",
        "bnuttall",
        "bhann",
        "sine",
        "nuttall",
        "lanczos",
        "gauss",
        "tukey",
        "dolph",
        "cauchy",
        "parzen",
        "poisson",
        "bohman",
        "kaiser",
    ]
    | int
    | None = None,
) -> "Stream":
    """Generate a FIR coefficients audio stream.
    Args:
        taps (int): set number of taps (from 9 to 65535)

            Defaults to 1025.
        t (int): set number of taps (from 9 to 65535)

            Defaults to 1025.
        frequency (str): set frequency points

            Defaults to 0 1.
        f (str): set frequency points

            Defaults to 0 1.
        magnitude (str): set magnitude values

            Defaults to 1 1.
        m (str): set magnitude values

            Defaults to 1 1.
        phase (str): set phase values

            Defaults to 0 0.
        p (str): set phase values

            Defaults to 0 0.
        sample_rate (int): set sample rate (from 1 to INT_MAX)

            Defaults to 44100.
        r (int): set sample rate (from 1 to INT_MAX)

            Defaults to 44100.
        nb_samples (int): set the number of samples per requested frame (from 1 to INT_MAX)

            Defaults to 1024.
        n (int): set the number of samples per requested frame (from 1 to INT_MAX)

            Defaults to 1024.
        win_func (int | str): set window function (from 0 to 20)

            Allowed values:
                * rect
                * bartlett
                * hann
                * hanning
                * hamming
                * blackman
                * welch
                * flattop
                * bharris
                * bnuttall
                * bhann
                * sine
                * nuttall
                * lanczos
                * gauss
                * tukey
                * dolph
                * cauchy
                * parzen
                * poisson
                * bohman
                * kaiser
                * rect
                * bartlett
                * hann
                * hanning
                * hamming
                * blackman
                * welch
                * flattop
                * bharris
                * bnuttall
                * bhann
                * sine
                * nuttall
                * lanczos
                * gauss
                * tukey
                * dolph
                * cauchy
                * parzen
                * poisson
                * bohman
                * kaiser

            Defaults to blackman.
        w (int | str): set window function (from 0 to 20)

            Allowed values:
                * rect
                * bartlett
                * hann
                * hanning
                * hamming
                * blackman
                * welch
                * flattop
                * bharris
                * bnuttall
                * bhann
                * sine
                * nuttall
                * lanczos
                * gauss
                * tukey
                * dolph
                * cauchy
                * parzen
                * poisson
                * bohman
                * kaiser
                * rect
                * bartlett
                * hann
                * hanning
                * hamming
                * blackman
                * welch
                * flattop
                * bharris
                * bnuttall
                * bhann
                * sine
                * nuttall
                * lanczos
                * gauss
                * tukey
                * dolph
                * cauchy
                * parzen
                * poisson
                * bohman
                * kaiser

            Defaults to blackman.

    Returns:
        "Stream": The output stream.
    """
    return create_source(
        filter_name="afirsrc",
        named_arguments={
            "taps": taps,
            "t": t,
            "frequency": frequency,
            "f": f,
            "magnitude": magnitude,
            "m": m,
            "phase": phase,
            "p": p,
            "sample_rate": sample_rate,
            "r": r,
            "nb_samples": nb_samples,
            "n": n,
            "win_func": win_func,
            "w": w,
        },
    )


def allrgb(
    rate: str | None = None,
    r: str | None = None,
    duration: str | None = None,
    d: str | None = None,
    sar: str | None = None,
) -> "Stream":
    """Generate all RGB colors.
    Args:
        rate (str): set video rate

            Defaults to 25.
        r (str): set video rate

            Defaults to 25.
        duration (str): set video duration

            Defaults to -0.000001.
        d (str): set video duration

            Defaults to -0.000001.
        sar (str): set video sample aspect ratio (from 0 to INT_MAX)

            Defaults to 1/1.

    Returns:
        "Stream": The output stream.
    """
    return create_source(
        filter_name="allrgb",
        named_arguments={
            "rate": rate,
            "r": r,
            "duration": duration,
            "d": d,
            "sar": sar,
        },
    )


def allyuv(
    rate: str | None = None,
    r: str | None = None,
    duration: str | None = None,
    d: str | None = None,
    sar: str | None = None,
) -> "Stream":
    """Generate all yuv colors.
    Args:
        rate (str): set video rate

            Defaults to 25.
        r (str): set video rate

            Defaults to 25.
        duration (str): set video duration

            Defaults to -0.000001.
        d (str): set video duration

            Defaults to -0.000001.
        sar (str): set video sample aspect ratio (from 0 to INT_MAX)

            Defaults to 1/1.

    Returns:
        "Stream": The output stream.
    """
    return create_source(
        filter_name="allyuv",
        named_arguments={
            "rate": rate,
            "r": r,
            "duration": duration,
            "d": d,
            "sar": sar,
        },
    )


def amovie(
    filename: str | None = None,
    format_name: str | None = None,
    f: str | None = None,
    stream_index: int | None = None,
    si: int | None = None,
    seek_point: float | None = None,
    sp: float | None = None,
    streams: str | None = None,
    s: str | None = None,
    loop: int | None = None,
    discontinuity: str | None = None,
    dec_threads: int | None = None,
    format_opts: str | None = None,
) -> "FilterMultiOutput":
    """Read audio from a movie source.
    Args:
        filename (str): No description available.

        format_name (str): set format name

        f (str): set format name

        stream_index (int): set stream index (from -1 to INT_MAX)

            Defaults to -1.
        si (int): set stream index (from -1 to INT_MAX)

            Defaults to -1.
        seek_point (float): set seekpoint (seconds) (from 0 to 9.22337e+12)

            Defaults to 0.
        sp (float): set seekpoint (seconds) (from 0 to 9.22337e+12)

            Defaults to 0.
        streams (str): set streams

        s (str): set streams

        loop (int): set loop count (from 0 to INT_MAX)

            Defaults to 1.
        discontinuity (str): set discontinuity threshold

            Defaults to 0.
        dec_threads (int): set the number of threads for decoding (from 0 to INT_MAX)

            Defaults to 0.
        format_opts (str): set format options for the opened file


    Returns:
        "FilterMultiOutput": A FilterMultiOutput object to access dynamic outputs.
    """
    return create_source(
        filter_name="amovie",
        named_arguments={
            "filename": filename,
            "format_name": format_name,
            "f": f,
            "stream_index": stream_index,
            "si": si,
            "seek_point": seek_point,
            "sp": sp,
            "streams": streams,
            "s": s,
            "loop": loop,
            "discontinuity": discontinuity,
            "dec_threads": dec_threads,
            "format_opts": format_opts,
        },
        is_dynamic=True,
    )


def anoisesrc(
    sample_rate: int | None = None,
    r: int | None = None,
    amplitude: float | None = None,
    a: float | None = None,
    duration: str | None = None,
    d: str | None = None,
    color: Literal["white", "pink", "brown", "blue", "violet", "velvet"]
    | int
    | None = None,
    colour: Literal["white", "pink", "brown", "blue", "violet", "velvet"]
    | int
    | None = None,
    c: Literal["white", "pink", "brown", "blue", "violet", "velvet"]
    | int
    | None = None,
    seed: str | None = None,
    s: str | None = None,
    nb_samples: int | None = None,
    n: int | None = None,
    density: float | None = None,
) -> "Stream":
    """Generate a noise audio signal.
    Args:
        sample_rate (int): set sample rate (from 15 to INT_MAX)

            Defaults to 48000.
        r (int): set sample rate (from 15 to INT_MAX)

            Defaults to 48000.
        amplitude (float): set amplitude (from 0 to 1)

            Defaults to 1.
        a (float): set amplitude (from 0 to 1)

            Defaults to 1.
        duration (str): set duration

            Defaults to 0.
        d (str): set duration

            Defaults to 0.
        color (int | str): set noise color (from 0 to 5)

            Allowed values:
                * white
                * pink
                * brown
                * blue
                * violet
                * velvet

            Defaults to white.
        colour (int | str): set noise color (from 0 to 5)

            Allowed values:
                * white
                * pink
                * brown
                * blue
                * violet
                * velvet

            Defaults to white.
        c (int | str): set noise color (from 0 to 5)

            Allowed values:
                * white
                * pink
                * brown
                * blue
                * violet
                * velvet

            Defaults to white.
        seed (str): set random seed (from -1 to UINT32_MAX)

            Defaults to -1.
        s (str): set random seed (from -1 to UINT32_MAX)

            Defaults to -1.
        nb_samples (int): set the number of samples per requested frame (from 1 to INT_MAX)

            Defaults to 1024.
        n (int): set the number of samples per requested frame (from 1 to INT_MAX)

            Defaults to 1024.
        density (float): set density (from 0 to 1)

            Defaults to 0.05.

    Returns:
        "Stream": The output stream.
    """
    return create_source(
        filter_name="anoisesrc",
        named_arguments={
            "sample_rate": sample_rate,
            "r": r,
            "amplitude": amplitude,
            "a": a,
            "duration": duration,
            "d": d,
            "color": color,
            "colour": colour,
            "c": c,
            "seed": seed,
            "s": s,
            "nb_samples": nb_samples,
            "n": n,
            "density": density,
        },
    )


def anullsrc(
    channel_layout: str | None = None,
    cl: str | None = None,
    sample_rate: int | None = None,
    r: int | None = None,
    nb_samples: int | None = None,
    n: int | None = None,
    duration: str | None = None,
    d: str | None = None,
) -> "Stream":
    """Null audio source, return empty audio frames.
    Args:
        channel_layout (str): set channel_layout

            Defaults to stereo.
        cl (str): set channel_layout

            Defaults to stereo.
        sample_rate (int): set sample rate (from 1 to INT_MAX)

            Defaults to 44100.
        r (int): set sample rate (from 1 to INT_MAX)

            Defaults to 44100.
        nb_samples (int): set the number of samples per requested frame (from 1 to 65535)

            Defaults to 1024.
        n (int): set the number of samples per requested frame (from 1 to 65535)

            Defaults to 1024.
        duration (str): set the audio duration

            Defaults to -0.000001.
        d (str): set the audio duration

            Defaults to -0.000001.

    Returns:
        "Stream": The output stream.
    """
    return create_source(
        filter_name="anullsrc",
        named_arguments={
            "channel_layout": channel_layout,
            "cl": cl,
            "sample_rate": sample_rate,
            "r": r,
            "nb_samples": nb_samples,
            "n": n,
            "duration": duration,
            "d": d,
        },
    )


def avsynctest(
    size: str | None = None,
    s: str | None = None,
    framerate: str | None = None,
    fr: str | None = None,
    samplerate: int | None = None,
    sr: int | None = None,
    amplitude: float | None = None,
    a: float | None = None,
    period: int | None = None,
    p: int | None = None,
    delay: int | None = None,
    dl: int | None = None,
    cycle: bool | None = None,
    c: bool | None = None,
    duration: str | None = None,
    d: str | None = None,
    fg: str | None = None,
    bg: str | None = None,
    ag: str | None = None,
) -> list["Stream"]:
    """Generate an Audio Video Sync Test.
    Args:
        size (str): set frame size

            Defaults to hd720.
        s (str): set frame size

            Defaults to hd720.
        framerate (str): set frame rate

            Defaults to 30.
        fr (str): set frame rate

            Defaults to 30.
        samplerate (int): set sample rate (from 8000 to 384000)

            Defaults to 44100.
        sr (int): set sample rate (from 8000 to 384000)

            Defaults to 44100.
        amplitude (float): set beep amplitude (from 0 to 1)

            Defaults to 0.7.
        a (float): set beep amplitude (from 0 to 1)

            Defaults to 0.7.
        period (int): set beep period (from 1 to 99)

            Defaults to 3.
        p (int): set beep period (from 1 to 99)

            Defaults to 3.
        delay (int): set flash delay (from -30 to 30)

            Defaults to 0.
        dl (int): set flash delay (from -30 to 30)

            Defaults to 0.
        cycle (bool): set delay cycle

            Defaults to false.
        c (bool): set delay cycle

            Defaults to false.
        duration (str): set duration

            Defaults to 0.
        d (str): set duration

            Defaults to 0.
        fg (str): set foreground color

            Defaults to white.
        bg (str): set background color

            Defaults to black.
        ag (str): set additional color

            Defaults to gray.

    Returns:
        list["Stream"]: A list of 2 Stream objects.
    """
    return create_source(
        filter_name="avsynctest",
        named_arguments={
            "size": size,
            "s": s,
            "framerate": framerate,
            "fr": fr,
            "samplerate": samplerate,
            "sr": sr,
            "amplitude": amplitude,
            "a": a,
            "period": period,
            "p": p,
            "delay": delay,
            "dl": dl,
            "cycle": cycle,
            "c": c,
            "duration": duration,
            "d": d,
            "fg": fg,
            "bg": bg,
            "ag": ag,
        },
        num_outputs=2,
    )


def buffer(
    width: int | None = None,
    video_size: str | None = None,
    height: int | None = None,
    pix_fmt: str | None = None,
    sar: str | None = None,
    pixel_aspect: str | None = None,
    time_base: str | None = None,
    frame_rate: str | None = None,
    colorspace: Literal[
        "gbr",
        "bt709",
        "unknown",
        "fcc",
        "bt470bg",
        "smpte170m",
        "smpte240m",
        "ycgco",
        "ycgco-re",
        "ycgco-ro",
        "bt2020nc",
        "bt2020c",
        "smpte2085",
        "chroma-derived-nc",
        "chroma-derived-c",
        "ictcp",
        "ipt-c2",
    ]
    | int
    | None = None,
    range: Literal[
        "unspecified", "unknown", "limited", "tv", "mpeg", "full", "pc", "jpeg"
    ]
    | int
    | None = None,
) -> "Stream":
    """Buffer video frames, and make them accessible to the filterchain.
    Args:
        width (int): (from 0 to INT_MAX)

            Defaults to 0.
        video_size (str): No description available.

        height (int): (from 0 to INT_MAX)

            Defaults to 0.
        pix_fmt (str): No description available.

            Defaults to none.
        sar (str): sample aspect ratio (from 0 to DBL_MAX)

            Defaults to 0/1.
        pixel_aspect (str): sample aspect ratio (from 0 to DBL_MAX)

            Defaults to 0/1.
        time_base (str): (from 0 to DBL_MAX)

            Defaults to 0/1.
        frame_rate (str): (from 0 to DBL_MAX)

            Defaults to 0/1.
        colorspace (int | str): select colorspace (from 0 to 17)

            Allowed values:
                * gbr
                * bt709
                * unknown
                * fcc
                * bt470bg
                * smpte170m
                * smpte240m
                * ycgco
                * ycgco-re
                * ycgco-ro
                * bt2020nc
                * bt2020c
                * smpte2085
                * chroma-derived-nc
                * chroma-derived-c
                * ictcp
                * ipt-c2

            Defaults to unknown.
        range (int | str): select color range (from 0 to 2)

            Allowed values:
                * unspecified
                * unknown
                * limited
                * tv
                * mpeg
                * full
                * pc
                * jpeg

            Defaults to unspecified.

    Returns:
        "Stream": The output stream.
    """
    return create_source(
        filter_name="buffer",
        named_arguments={
            "width": width,
            "video_size": video_size,
            "height": height,
            "pix_fmt": pix_fmt,
            "sar": sar,
            "pixel_aspect": pixel_aspect,
            "time_base": time_base,
            "frame_rate": frame_rate,
            "colorspace": colorspace,
            "range": range,
        },
    )


def cellauto(
    filename: str | None = None,
    f: str | None = None,
    pattern: str | None = None,
    p: str | None = None,
    rate: str | None = None,
    r: str | None = None,
    size: str | None = None,
    s: str | None = None,
    rule: int | None = None,
    random_fill_ratio: float | None = None,
    ratio: float | None = None,
    random_seed: str | None = None,
    seed: str | None = None,
    scroll: bool | None = None,
    start_full: bool | None = None,
    full: bool | None = None,
    stitch: bool | None = None,
) -> "Stream":
    """Create pattern generated by an elementary cellular automaton.
    Args:
        filename (str): read initial pattern from file

        f (str): read initial pattern from file

        pattern (str): set initial pattern

        p (str): set initial pattern

        rate (str): set video rate

            Defaults to 25.
        r (str): set video rate

            Defaults to 25.
        size (str): set video size

        s (str): set video size

        rule (int): set rule (from 0 to 255)

            Defaults to 110.
        random_fill_ratio (float): set fill ratio for filling initial grid randomly (from 0 to 1)

            Defaults to 0.618034.
        ratio (float): set fill ratio for filling initial grid randomly (from 0 to 1)

            Defaults to 0.618034.
        random_seed (str): set the seed for filling the initial grid randomly (from -1 to UINT32_MAX)

            Defaults to -1.
        seed (str): set the seed for filling the initial grid randomly (from -1 to UINT32_MAX)

            Defaults to -1.
        scroll (bool): scroll pattern downward

            Defaults to true.
        start_full (bool): start filling the whole video

            Defaults to false.
        full (bool): start filling the whole video

            Defaults to true.
        stitch (bool): stitch boundaries

            Defaults to true.

    Returns:
        "Stream": The output stream.
    """
    return create_source(
        filter_name="cellauto",
        named_arguments={
            "filename": filename,
            "f": f,
            "pattern": pattern,
            "p": p,
            "rate": rate,
            "r": r,
            "size": size,
            "s": s,
            "rule": rule,
            "random_fill_ratio": random_fill_ratio,
            "ratio": ratio,
            "random_seed": random_seed,
            "seed": seed,
            "scroll": scroll,
            "start_full": start_full,
            "full": full,
            "stitch": stitch,
        },
    )


def color(
    color: str | None = None,
    c: str | None = None,
    size: str | None = None,
    s: str | None = None,
    rate: str | None = None,
    r: str | None = None,
    duration: str | None = None,
    d: str | None = None,
    sar: str | None = None,
) -> "Stream":
    """Provide an uniformly colored input.
    Args:
        color (str): set color

            Defaults to black.
        c (str): set color

            Defaults to black.
        size (str): set video size

            Defaults to 320x240.
        s (str): set video size

            Defaults to 320x240.
        rate (str): set video rate

            Defaults to 25.
        r (str): set video rate

            Defaults to 25.
        duration (str): set video duration

            Defaults to -0.000001.
        d (str): set video duration

            Defaults to -0.000001.
        sar (str): set video sample aspect ratio (from 0 to INT_MAX)

            Defaults to 1/1.

    Returns:
        "Stream": The output stream.
    """
    return create_source(
        filter_name="color",
        named_arguments={
            "color": color,
            "c": c,
            "size": size,
            "s": s,
            "rate": rate,
            "r": r,
            "duration": duration,
            "d": d,
            "sar": sar,
        },
    )


def colorchart(
    rate: str | None = None,
    r: str | None = None,
    duration: str | None = None,
    d: str | None = None,
    sar: str | None = None,
    patch_size: str | None = None,
    preset: Literal["reference", "skintones"] | int | None = None,
) -> "Stream":
    """Generate color checker chart.
    Args:
        rate (str): set video rate

            Defaults to 25.
        r (str): set video rate

            Defaults to 25.
        duration (str): set video duration

            Defaults to -0.000001.
        d (str): set video duration

            Defaults to -0.000001.
        sar (str): set video sample aspect ratio (from 0 to INT_MAX)

            Defaults to 1/1.
        patch_size (str): set the single patch size

            Defaults to 64x64.
        preset (int | str): set the color checker chart preset (from 0 to 1)

            Allowed values:
                * reference
                * skintones

            Defaults to reference.

    Returns:
        "Stream": The output stream.
    """
    return create_source(
        filter_name="colorchart",
        named_arguments={
            "rate": rate,
            "r": r,
            "duration": duration,
            "d": d,
            "sar": sar,
            "patch_size": patch_size,
            "preset": preset,
        },
    )


def colorspectrum(
    size: str | None = None,
    s: str | None = None,
    rate: str | None = None,
    r: str | None = None,
    duration: str | None = None,
    d: str | None = None,
    sar: str | None = None,
    type: Literal["black", "white", "all"] | int | None = None,
) -> "Stream":
    """Generate colors spectrum.
    Args:
        size (str): set video size

            Defaults to 320x240.
        s (str): set video size

            Defaults to 320x240.
        rate (str): set video rate

            Defaults to 25.
        r (str): set video rate

            Defaults to 25.
        duration (str): set video duration

            Defaults to -0.000001.
        d (str): set video duration

            Defaults to -0.000001.
        sar (str): set video sample aspect ratio (from 0 to INT_MAX)

            Defaults to 1/1.
        type (int | str): set the color spectrum type (from 0 to 2)

            Allowed values:
                * black
                * white
                * all

            Defaults to black.

    Returns:
        "Stream": The output stream.
    """
    return create_source(
        filter_name="colorspectrum",
        named_arguments={
            "size": size,
            "s": s,
            "rate": rate,
            "r": r,
            "duration": duration,
            "d": d,
            "sar": sar,
            "type": type,
        },
    )


def coreimagesrc(
    size: str | None = None,
    s: str | None = None,
    rate: str | None = None,
    r: str | None = None,
    duration: str | None = None,
    d: str | None = None,
    sar: str | None = None,
    list_filters: bool | None = None,
    list_generators: bool | None = None,
    filter: str | None = None,
    output_rect: str | None = None,
) -> "Stream":
    """Video source using image generators of CoreImage API.
    Args:
        size (str): set video size

            Defaults to 320x240.
        s (str): set video size

            Defaults to 320x240.
        rate (str): set video rate

            Defaults to 25.
        r (str): set video rate

            Defaults to 25.
        duration (str): set video duration

            Defaults to -0.000001.
        d (str): set video duration

            Defaults to -0.000001.
        sar (str): set video sample aspect ratio (from 0 to INT_MAX)

            Defaults to 1/1.
        list_filters (bool): list available filters

            Defaults to false.
        list_generators (bool): list available generators

            Defaults to false.
        filter (str): names and options of filters to apply

        output_rect (str): output rectangle within output image


    Returns:
        "Stream": The output stream.
    """
    return create_source(
        filter_name="coreimagesrc",
        named_arguments={
            "size": size,
            "s": s,
            "rate": rate,
            "r": r,
            "duration": duration,
            "d": d,
            "sar": sar,
            "list_filters": list_filters,
            "list_generators": list_generators,
            "filter": filter,
            "output_rect": output_rect,
        },
    )


def frei0r_src(
    size: str | None = None,
    framerate: str | None = None,
    filter_name: str | None = None,
    filter_params: str | None = None,
) -> "Stream":
    """Generate a frei0r source.
    Args:
        size (str): Dimensions of the generated video.

            Defaults to 320x240.
        framerate (str): No description available.

            Defaults to 25.
        filter_name (str): No description available.

        filter_params (str): No description available.


    Returns:
        "Stream": The output stream.
    """
    return create_source(
        filter_name="frei0r_src",
        named_arguments={
            "size": size,
            "framerate": framerate,
            "filter_name": filter_name,
            "filter_params": filter_params,
        },
    )


def gradients(
    size: str | None = None,
    s: str | None = None,
    rate: str | None = None,
    r: str | None = None,
    c0: str | None = None,
    c1: str | None = None,
    c2: str | None = None,
    c3: str | None = None,
    c4: str | None = None,
    c5: str | None = None,
    c6: str | None = None,
    c7: str | None = None,
    x0: int | None = None,
    y0: int | None = None,
    x1: int | None = None,
    y1: int | None = None,
    nb_colors: int | None = None,
    n: int | None = None,
    seed: str | None = None,
    duration: str | None = None,
    d: str | None = None,
    speed: float | None = None,
    type: Literal["linear", "radial", "circular", "spiral", "square"]
    | int
    | None = None,
    t: Literal["linear", "radial", "circular", "spiral", "square"] | int | None = None,
) -> "Stream":
    """Draw a gradients.
    Args:
        size (str): set frame size

            Defaults to 640x480.
        s (str): set frame size

            Defaults to 640x480.
        rate (str): set frame rate

            Defaults to 25.
        r (str): set frame rate

            Defaults to 25.
        c0 (str): set 1st color

            Defaults to random.
        c1 (str): set 2nd color

            Defaults to random.
        c2 (str): set 3rd color

            Defaults to random.
        c3 (str): set 4th color

            Defaults to random.
        c4 (str): set 5th color

            Defaults to random.
        c5 (str): set 6th color

            Defaults to random.
        c6 (str): set 7th color

            Defaults to random.
        c7 (str): set 8th color

            Defaults to random.
        x0 (int): set gradient line source x0 (from -1 to INT_MAX)

            Defaults to -1.
        y0 (int): set gradient line source y0 (from -1 to INT_MAX)

            Defaults to -1.
        x1 (int): set gradient line destination x1 (from -1 to INT_MAX)

            Defaults to -1.
        y1 (int): set gradient line destination y1 (from -1 to INT_MAX)

            Defaults to -1.
        nb_colors (int): set the number of colors (from 2 to 8)

            Defaults to 2.
        n (int): set the number of colors (from 2 to 8)

            Defaults to 2.
        seed (str): set the seed (from -1 to UINT32_MAX)

            Defaults to -1.
        duration (str): set video duration

            Defaults to -0.000001.
        d (str): set video duration

            Defaults to -0.000001.
        speed (float): set gradients rotation speed (from 0 to 1)

            Defaults to 0.01.
        type (int | str): set gradient type (from 0 to 4)

            Allowed values:
                * linear
                * radial
                * circular
                * spiral
                * square

            Defaults to linear.
        t (int | str): set gradient type (from 0 to 4)

            Allowed values:
                * linear
                * radial
                * circular
                * spiral
                * square

            Defaults to linear.

    Returns:
        "Stream": The output stream.
    """
    return create_source(
        filter_name="gradients",
        named_arguments={
            "size": size,
            "s": s,
            "rate": rate,
            "r": r,
            "c0": c0,
            "c1": c1,
            "c2": c2,
            "c3": c3,
            "c4": c4,
            "c5": c5,
            "c6": c6,
            "c7": c7,
            "x0": x0,
            "y0": y0,
            "x1": x1,
            "y1": y1,
            "nb_colors": nb_colors,
            "n": n,
            "seed": seed,
            "duration": duration,
            "d": d,
            "speed": speed,
            "type": type,
            "t": t,
        },
    )


def haldclutsrc(
    level: int | None = None,
    rate: str | None = None,
    r: str | None = None,
    duration: str | None = None,
    d: str | None = None,
    sar: str | None = None,
) -> "Stream":
    """Provide an identity Hald CLUT.
    Args:
        level (int): set level (from 2 to 16)

            Defaults to 6.
        rate (str): set video rate

            Defaults to 25.
        r (str): set video rate

            Defaults to 25.
        duration (str): set video duration

            Defaults to -0.000001.
        d (str): set video duration

            Defaults to -0.000001.
        sar (str): set video sample aspect ratio (from 0 to INT_MAX)

            Defaults to 1/1.

    Returns:
        "Stream": The output stream.
    """
    return create_source(
        filter_name="haldclutsrc",
        named_arguments={
            "level": level,
            "rate": rate,
            "r": r,
            "duration": duration,
            "d": d,
            "sar": sar,
        },
    )


def hilbert(
    sample_rate: int | None = None,
    r: int | None = None,
    taps: int | None = None,
    t: int | None = None,
    nb_samples: int | None = None,
    n: int | None = None,
    win_func: Literal[
        "rect",
        "bartlett",
        "hann",
        "hanning",
        "hamming",
        "blackman",
        "welch",
        "flattop",
        "bharris",
        "bnuttall",
        "bhann",
        "sine",
        "nuttall",
        "lanczos",
        "gauss",
        "tukey",
        "dolph",
        "cauchy",
        "parzen",
        "poisson",
        "bohman",
        "kaiser",
        "rect",
        "bartlett",
        "hann",
        "hanning",
        "hamming",
        "blackman",
        "welch",
        "flattop",
        "bharris",
        "bnuttall",
        "bhann",
        "sine",
        "nuttall",
        "lanczos",
        "gauss",
        "tukey",
        "dolph",
        "cauchy",
        "parzen",
        "poisson",
        "bohman",
        "kaiser",
    ]
    | int
    | None = None,
    w: Literal[
        "rect",
        "bartlett",
        "hann",
        "hanning",
        "hamming",
        "blackman",
        "welch",
        "flattop",
        "bharris",
        "bnuttall",
        "bhann",
        "sine",
        "nuttall",
        "lanczos",
        "gauss",
        "tukey",
        "dolph",
        "cauchy",
        "parzen",
        "poisson",
        "bohman",
        "kaiser",
        "rect",
        "bartlett",
        "hann",
        "hanning",
        "hamming",
        "blackman",
        "welch",
        "flattop",
        "bharris",
        "bnuttall",
        "bhann",
        "sine",
        "nuttall",
        "lanczos",
        "gauss",
        "tukey",
        "dolph",
        "cauchy",
        "parzen",
        "poisson",
        "bohman",
        "kaiser",
    ]
    | int
    | None = None,
) -> "Stream":
    """Generate a Hilbert transform FIR coefficients.
    Args:
        sample_rate (int): set sample rate (from 1 to INT_MAX)

            Defaults to 44100.
        r (int): set sample rate (from 1 to INT_MAX)

            Defaults to 44100.
        taps (int): set number of taps (from 11 to 65535)

            Defaults to 22051.
        t (int): set number of taps (from 11 to 65535)

            Defaults to 22051.
        nb_samples (int): set the number of samples per requested frame (from 1 to INT_MAX)

            Defaults to 1024.
        n (int): set the number of samples per requested frame (from 1 to INT_MAX)

            Defaults to 1024.
        win_func (int | str): set window function (from 0 to 20)

            Allowed values:
                * rect
                * bartlett
                * hann
                * hanning
                * hamming
                * blackman
                * welch
                * flattop
                * bharris
                * bnuttall
                * bhann
                * sine
                * nuttall
                * lanczos
                * gauss
                * tukey
                * dolph
                * cauchy
                * parzen
                * poisson
                * bohman
                * kaiser
                * rect
                * bartlett
                * hann
                * hanning
                * hamming
                * blackman
                * welch
                * flattop
                * bharris
                * bnuttall
                * bhann
                * sine
                * nuttall
                * lanczos
                * gauss
                * tukey
                * dolph
                * cauchy
                * parzen
                * poisson
                * bohman
                * kaiser

            Defaults to blackman.
        w (int | str): set window function (from 0 to 20)

            Allowed values:
                * rect
                * bartlett
                * hann
                * hanning
                * hamming
                * blackman
                * welch
                * flattop
                * bharris
                * bnuttall
                * bhann
                * sine
                * nuttall
                * lanczos
                * gauss
                * tukey
                * dolph
                * cauchy
                * parzen
                * poisson
                * bohman
                * kaiser
                * rect
                * bartlett
                * hann
                * hanning
                * hamming
                * blackman
                * welch
                * flattop
                * bharris
                * bnuttall
                * bhann
                * sine
                * nuttall
                * lanczos
                * gauss
                * tukey
                * dolph
                * cauchy
                * parzen
                * poisson
                * bohman
                * kaiser

            Defaults to blackman.

    Returns:
        "Stream": The output stream.
    """
    return create_source(
        filter_name="hilbert",
        named_arguments={
            "sample_rate": sample_rate,
            "r": r,
            "taps": taps,
            "t": t,
            "nb_samples": nb_samples,
            "n": n,
            "win_func": win_func,
            "w": w,
        },
    )


def life(
    filename: str | None = None,
    f: str | None = None,
    size: str | None = None,
    s: str | None = None,
    rate: str | None = None,
    r: str | None = None,
    rule: str | None = None,
    random_fill_ratio: float | None = None,
    ratio: float | None = None,
    random_seed: str | None = None,
    seed: str | None = None,
    stitch: bool | None = None,
    mold: int | None = None,
    life_color: str | None = None,
    death_color: str | None = None,
    mold_color: str | None = None,
) -> "Stream":
    """Create life.
    Args:
        filename (str): set source file

        f (str): set source file

        size (str): set video size

        s (str): set video size

        rate (str): set video rate

            Defaults to 25.
        r (str): set video rate

            Defaults to 25.
        rule (str): set rule

            Defaults to B3/S23.
        random_fill_ratio (float): set fill ratio for filling initial grid randomly (from 0 to 1)

            Defaults to 0.618034.
        ratio (float): set fill ratio for filling initial grid randomly (from 0 to 1)

            Defaults to 0.618034.
        random_seed (str): set the seed for filling the initial grid randomly (from -1 to UINT32_MAX)

            Defaults to -1.
        seed (str): set the seed for filling the initial grid randomly (from -1 to UINT32_MAX)

            Defaults to -1.
        stitch (bool): stitch boundaries

            Defaults to true.
        mold (int): set mold speed for dead cells (from 0 to 255)

            Defaults to 0.
        life_color (str): set life color

            Defaults to white.
        death_color (str): set death color

            Defaults to black.
        mold_color (str): set mold color

            Defaults to black.

    Returns:
        "Stream": The output stream.
    """
    return create_source(
        filter_name="life",
        named_arguments={
            "filename": filename,
            "f": f,
            "size": size,
            "s": s,
            "rate": rate,
            "r": r,
            "rule": rule,
            "random_fill_ratio": random_fill_ratio,
            "ratio": ratio,
            "random_seed": random_seed,
            "seed": seed,
            "stitch": stitch,
            "mold": mold,
            "life_color": life_color,
            "death_color": death_color,
            "mold_color": mold_color,
        },
    )


def mandelbrot(
    size: str | None = None,
    s: str | None = None,
    rate: str | None = None,
    r: str | None = None,
    maxiter: int | None = None,
    start_x: float | None = None,
    start_y: float | None = None,
    start_scale: float | None = None,
    end_scale: float | None = None,
    end_pts: float | None = None,
    bailout: float | None = None,
    morphxf: float | None = None,
    morphyf: float | None = None,
    morphamp: float | None = None,
    outer: Literal["iteration_count", "normalized_iteration_count", "white", "outz"]
    | int
    | None = None,
    inner: Literal["black", "period", "convergence", "mincol"] | int | None = None,
) -> "Stream":
    """Render a Mandelbrot fractal.
    Args:
        size (str): set frame size

            Defaults to 640x480.
        s (str): set frame size

            Defaults to 640x480.
        rate (str): set frame rate

            Defaults to 25.
        r (str): set frame rate

            Defaults to 25.
        maxiter (int): set max iterations number (from 1 to INT_MAX)

            Defaults to 7189.
        start_x (float): set the initial x position (from -100 to 100)

            Defaults to -0.743644.
        start_y (float): set the initial y position (from -100 to 100)

            Defaults to -0.131826.
        start_scale (float): set the initial scale value (from 0 to FLT_MAX)

            Defaults to 3.
        end_scale (float): set the terminal scale value (from 0 to FLT_MAX)

            Defaults to 0.3.
        end_pts (float): set the terminal pts value (from 0 to I64_MAX)

            Defaults to 400.
        bailout (float): set the bailout value (from 0 to FLT_MAX)

            Defaults to 10.
        morphxf (float): set morph x frequency (from -FLT_MAX to FLT_MAX)

            Defaults to 0.01.
        morphyf (float): set morph y frequency (from -FLT_MAX to FLT_MAX)

            Defaults to 0.0123.
        morphamp (float): set morph amplitude (from -FLT_MAX to FLT_MAX)

            Defaults to 0.
        outer (int | str): set outer coloring mode (from 0 to INT_MAX)

            Allowed values:
                * iteration_count
                * normalized_iteration_count
                * white
                * outz

            Defaults to normalized_iteration_count.
        inner (int | str): set inner coloring mode (from 0 to INT_MAX)

            Allowed values:
                * black
                * period
                * convergence
                * mincol

            Defaults to mincol.

    Returns:
        "Stream": The output stream.
    """
    return create_source(
        filter_name="mandelbrot",
        named_arguments={
            "size": size,
            "s": s,
            "rate": rate,
            "r": r,
            "maxiter": maxiter,
            "start_x": start_x,
            "start_y": start_y,
            "start_scale": start_scale,
            "end_scale": end_scale,
            "end_pts": end_pts,
            "bailout": bailout,
            "morphxf": morphxf,
            "morphyf": morphyf,
            "morphamp": morphamp,
            "outer": outer,
            "inner": inner,
        },
    )


def movie(
    filename: str | None = None,
    format_name: str | None = None,
    f: str | None = None,
    stream_index: int | None = None,
    si: int | None = None,
    seek_point: float | None = None,
    sp: float | None = None,
    streams: str | None = None,
    s: str | None = None,
    loop: int | None = None,
    discontinuity: str | None = None,
    dec_threads: int | None = None,
    format_opts: str | None = None,
) -> "FilterMultiOutput":
    """Read from a movie source.
    Args:
        filename (str): No description available.

        format_name (str): set format name

        f (str): set format name

        stream_index (int): set stream index (from -1 to INT_MAX)

            Defaults to -1.
        si (int): set stream index (from -1 to INT_MAX)

            Defaults to -1.
        seek_point (float): set seekpoint (seconds) (from 0 to 9.22337e+12)

            Defaults to 0.
        sp (float): set seekpoint (seconds) (from 0 to 9.22337e+12)

            Defaults to 0.
        streams (str): set streams

        s (str): set streams

        loop (int): set loop count (from 0 to INT_MAX)

            Defaults to 1.
        discontinuity (str): set discontinuity threshold

            Defaults to 0.
        dec_threads (int): set the number of threads for decoding (from 0 to INT_MAX)

            Defaults to 0.
        format_opts (str): set format options for the opened file


    Returns:
        "FilterMultiOutput": A FilterMultiOutput object to access dynamic outputs.
    """
    return create_source(
        filter_name="movie",
        named_arguments={
            "filename": filename,
            "format_name": format_name,
            "f": f,
            "stream_index": stream_index,
            "si": si,
            "seek_point": seek_point,
            "sp": sp,
            "streams": streams,
            "s": s,
            "loop": loop,
            "discontinuity": discontinuity,
            "dec_threads": dec_threads,
            "format_opts": format_opts,
        },
        is_dynamic=True,
    )


def mptestsrc(
    rate: str | None = None,
    r: str | None = None,
    duration: str | None = None,
    d: str | None = None,
    test: Literal[
        "dc_luma",
        "dc_chroma",
        "freq_luma",
        "freq_chroma",
        "amp_luma",
        "amp_chroma",
        "cbp",
        "mv",
        "ring1",
        "ring2",
        "all",
    ]
    | int
    | None = None,
    t: Literal[
        "dc_luma",
        "dc_chroma",
        "freq_luma",
        "freq_chroma",
        "amp_luma",
        "amp_chroma",
        "cbp",
        "mv",
        "ring1",
        "ring2",
        "all",
    ]
    | int
    | None = None,
    max_frames: str | None = None,
    m: str | None = None,
) -> "Stream":
    """Generate various test pattern.
    Args:
        rate (str): set video rate

            Defaults to 25.
        r (str): set video rate

            Defaults to 25.
        duration (str): set video duration

            Defaults to -0.000001.
        d (str): set video duration

            Defaults to -0.000001.
        test (int | str): set test to perform (from 0 to INT_MAX)

            Allowed values:
                * dc_luma
                * dc_chroma
                * freq_luma
                * freq_chroma
                * amp_luma
                * amp_chroma
                * cbp
                * mv
                * ring1
                * ring2
                * all

            Defaults to all.
        t (int | str): set test to perform (from 0 to INT_MAX)

            Allowed values:
                * dc_luma
                * dc_chroma
                * freq_luma
                * freq_chroma
                * amp_luma
                * amp_chroma
                * cbp
                * mv
                * ring1
                * ring2
                * all

            Defaults to all.
        max_frames (str): Set the maximum number of frames generated for each test (from 1 to I64_MAX)

            Defaults to 30.
        m (str): Set the maximum number of frames generated for each test (from 1 to I64_MAX)

            Defaults to 30.

    Returns:
        "Stream": The output stream.
    """
    return create_source(
        filter_name="mptestsrc",
        named_arguments={
            "rate": rate,
            "r": r,
            "duration": duration,
            "d": d,
            "test": test,
            "t": t,
            "max_frames": max_frames,
            "m": m,
        },
    )


def nullsrc(
    size: str | None = None,
    s: str | None = None,
    rate: str | None = None,
    r: str | None = None,
    duration: str | None = None,
    d: str | None = None,
    sar: str | None = None,
) -> "Stream":
    """Null video source, return unprocessed video frames.
    Args:
        size (str): set video size

            Defaults to 320x240.
        s (str): set video size

            Defaults to 320x240.
        rate (str): set video rate

            Defaults to 25.
        r (str): set video rate

            Defaults to 25.
        duration (str): set video duration

            Defaults to -0.000001.
        d (str): set video duration

            Defaults to -0.000001.
        sar (str): set video sample aspect ratio (from 0 to INT_MAX)

            Defaults to 1/1.

    Returns:
        "Stream": The output stream.
    """
    return create_source(
        filter_name="nullsrc",
        named_arguments={
            "size": size,
            "s": s,
            "rate": rate,
            "r": r,
            "duration": duration,
            "d": d,
            "sar": sar,
        },
    )


def pal100bars(
    size: str | None = None,
    s: str | None = None,
    rate: str | None = None,
    r: str | None = None,
    duration: str | None = None,
    d: str | None = None,
    sar: str | None = None,
) -> "Stream":
    """Generate PAL 100% color bars.
    Args:
        size (str): set video size

            Defaults to 320x240.
        s (str): set video size

            Defaults to 320x240.
        rate (str): set video rate

            Defaults to 25.
        r (str): set video rate

            Defaults to 25.
        duration (str): set video duration

            Defaults to -0.000001.
        d (str): set video duration

            Defaults to -0.000001.
        sar (str): set video sample aspect ratio (from 0 to INT_MAX)

            Defaults to 1/1.

    Returns:
        "Stream": The output stream.
    """
    return create_source(
        filter_name="pal100bars",
        named_arguments={
            "size": size,
            "s": s,
            "rate": rate,
            "r": r,
            "duration": duration,
            "d": d,
            "sar": sar,
        },
    )


def pal75bars(
    size: str | None = None,
    s: str | None = None,
    rate: str | None = None,
    r: str | None = None,
    duration: str | None = None,
    d: str | None = None,
    sar: str | None = None,
) -> "Stream":
    """Generate PAL 75% color bars.
    Args:
        size (str): set video size

            Defaults to 320x240.
        s (str): set video size

            Defaults to 320x240.
        rate (str): set video rate

            Defaults to 25.
        r (str): set video rate

            Defaults to 25.
        duration (str): set video duration

            Defaults to -0.000001.
        d (str): set video duration

            Defaults to -0.000001.
        sar (str): set video sample aspect ratio (from 0 to INT_MAX)

            Defaults to 1/1.

    Returns:
        "Stream": The output stream.
    """
    return create_source(
        filter_name="pal75bars",
        named_arguments={
            "size": size,
            "s": s,
            "rate": rate,
            "r": r,
            "duration": duration,
            "d": d,
            "sar": sar,
        },
    )


def perlin(
    size: str | None = None,
    s: str | None = None,
    rate: str | None = None,
    r: str | None = None,
    octaves: int | None = None,
    persistence: float | None = None,
    xscale: float | None = None,
    yscale: float | None = None,
    tscale: float | None = None,
    random_mode: Literal["random", "ken", "seed"] | int | None = None,
    random_seed: str | None = None,
    seed: str | None = None,
) -> "Stream":
    """Generate Perlin noise
    Args:
        size (str): set video size

            Defaults to 320x240.
        s (str): set video size

            Defaults to 320x240.
        rate (str): set video rate

            Defaults to 25.
        r (str): set video rate

            Defaults to 25.
        octaves (int): set the number of components to use to generate the noise (from 1 to INT_MAX)

            Defaults to 1.
        persistence (float): set the octaves persistence (from 0 to DBL_MAX)

            Defaults to 1.
        xscale (float): set x-scale factor (from 0 to DBL_MAX)

            Defaults to 1.
        yscale (float): set y-scale factor (from 0 to DBL_MAX)

            Defaults to 1.
        tscale (float): set t-scale factor (from 0 to DBL_MAX)

            Defaults to 1.
        random_mode (int | str): set random mode used to compute initial pattern (from 0 to 2)

            Allowed values:
                * random
                * ken
                * seed

            Defaults to random.
        random_seed (str): set the seed for filling the initial pattern (from 0 to UINT32_MAX)

            Defaults to 0.
        seed (str): set the seed for filling the initial pattern (from 0 to UINT32_MAX)

            Defaults to 0.

    Returns:
        "Stream": The output stream.
    """
    return create_source(
        filter_name="perlin",
        named_arguments={
            "size": size,
            "s": s,
            "rate": rate,
            "r": r,
            "octaves": octaves,
            "persistence": persistence,
            "xscale": xscale,
            "yscale": yscale,
            "tscale": tscale,
            "random_mode": random_mode,
            "random_seed": random_seed,
            "seed": seed,
        },
    )


def rgbtestsrc(
    size: str | None = None,
    s: str | None = None,
    rate: str | None = None,
    r: str | None = None,
    duration: str | None = None,
    d: str | None = None,
    sar: str | None = None,
    complement: bool | None = None,
    co: bool | None = None,
) -> "Stream":
    """Generate RGB test pattern.
    Args:
        size (str): set video size

            Defaults to 320x240.
        s (str): set video size

            Defaults to 320x240.
        rate (str): set video rate

            Defaults to 25.
        r (str): set video rate

            Defaults to 25.
        duration (str): set video duration

            Defaults to -0.000001.
        d (str): set video duration

            Defaults to -0.000001.
        sar (str): set video sample aspect ratio (from 0 to INT_MAX)

            Defaults to 1/1.
        complement (bool): set complement colors

            Defaults to false.
        co (bool): set complement colors

            Defaults to false.

    Returns:
        "Stream": The output stream.
    """
    return create_source(
        filter_name="rgbtestsrc",
        named_arguments={
            "size": size,
            "s": s,
            "rate": rate,
            "r": r,
            "duration": duration,
            "d": d,
            "sar": sar,
            "complement": complement,
            "co": co,
        },
    )


def sierpinski(
    size: str | None = None,
    s: str | None = None,
    rate: str | None = None,
    r: str | None = None,
    seed: str | None = None,
    jump: int | None = None,
    type: Literal["carpet", "triangle"] | int | None = None,
) -> "Stream":
    """Render a Sierpinski fractal.
    Args:
        size (str): set frame size

            Defaults to 640x480.
        s (str): set frame size

            Defaults to 640x480.
        rate (str): set frame rate

            Defaults to 25.
        r (str): set frame rate

            Defaults to 25.
        seed (str): set the seed (from -1 to UINT32_MAX)

            Defaults to -1.
        jump (int): set the jump (from 1 to 10000)

            Defaults to 100.
        type (int | str): set fractal type (from 0 to 1)

            Allowed values:
                * carpet
                * triangle

            Defaults to carpet.

    Returns:
        "Stream": The output stream.
    """
    return create_source(
        filter_name="sierpinski",
        named_arguments={
            "size": size,
            "s": s,
            "rate": rate,
            "r": r,
            "seed": seed,
            "jump": jump,
            "type": type,
        },
    )


def sinc(
    sample_rate: int | None = None,
    r: int | None = None,
    nb_samples: int | None = None,
    n: int | None = None,
    hp: float | None = None,
    lp: float | None = None,
    phase: float | None = None,
    beta: float | None = None,
    att: float | None = None,
    round: bool | None = None,
    hptaps: int | None = None,
    lptaps: int | None = None,
) -> "Stream":
    """Generate a sinc kaiser-windowed low-pass, high-pass, band-pass, or band-reject FIR coefficients.
    Args:
        sample_rate (int): set sample rate (from 1 to INT_MAX)

            Defaults to 44100.
        r (int): set sample rate (from 1 to INT_MAX)

            Defaults to 44100.
        nb_samples (int): set the number of samples per requested frame (from 1 to INT_MAX)

            Defaults to 1024.
        n (int): set the number of samples per requested frame (from 1 to INT_MAX)

            Defaults to 1024.
        hp (float): set high-pass filter frequency (from 0 to INT_MAX)

            Defaults to 0.
        lp (float): set low-pass filter frequency (from 0 to INT_MAX)

            Defaults to 0.
        phase (float): set filter phase response (from 0 to 100)

            Defaults to 50.
        beta (float): set kaiser window beta (from -1 to 256)

            Defaults to -1.
        att (float): set stop-band attenuation (from 40 to 180)

            Defaults to 120.
        round (bool): enable rounding

            Defaults to false.
        hptaps (int): set number of taps for high-pass filter (from 0 to 32768)

            Defaults to 0.
        lptaps (int): set number of taps for low-pass filter (from 0 to 32768)

            Defaults to 0.

    Returns:
        "Stream": The output stream.
    """
    return create_source(
        filter_name="sinc",
        named_arguments={
            "sample_rate": sample_rate,
            "r": r,
            "nb_samples": nb_samples,
            "n": n,
            "hp": hp,
            "lp": lp,
            "phase": phase,
            "beta": beta,
            "att": att,
            "round": round,
            "hptaps": hptaps,
            "lptaps": lptaps,
        },
    )


def sine(
    frequency: float | None = None,
    f: float | None = None,
    beep_factor: float | None = None,
    b: float | None = None,
    sample_rate: int | None = None,
    r: int | None = None,
    duration: str | None = None,
    d: str | None = None,
    samples_per_frame: str | None = None,
) -> "Stream":
    """Generate sine wave audio signal.
    Args:
        frequency (float): set the sine frequency (from 0 to DBL_MAX)

            Defaults to 440.
        f (float): set the sine frequency (from 0 to DBL_MAX)

            Defaults to 440.
        beep_factor (float): set the beep frequency factor (from 0 to DBL_MAX)

            Defaults to 0.
        b (float): set the beep frequency factor (from 0 to DBL_MAX)

            Defaults to 0.
        sample_rate (int): set the sample rate (from 1 to INT_MAX)

            Defaults to 44100.
        r (int): set the sample rate (from 1 to INT_MAX)

            Defaults to 44100.
        duration (str): set the audio duration

            Defaults to 0.
        d (str): set the audio duration

            Defaults to 0.
        samples_per_frame (str): set the number of samples per frame

            Defaults to 1024.

    Returns:
        "Stream": The output stream.
    """
    return create_source(
        filter_name="sine",
        named_arguments={
            "frequency": frequency,
            "f": f,
            "beep_factor": beep_factor,
            "b": b,
            "sample_rate": sample_rate,
            "r": r,
            "duration": duration,
            "d": d,
            "samples_per_frame": samples_per_frame,
        },
    )


def smptebars(
    size: str | None = None,
    s: str | None = None,
    rate: str | None = None,
    r: str | None = None,
    duration: str | None = None,
    d: str | None = None,
    sar: str | None = None,
) -> "Stream":
    """Generate SMPTE color bars.
    Args:
        size (str): set video size

            Defaults to 320x240.
        s (str): set video size

            Defaults to 320x240.
        rate (str): set video rate

            Defaults to 25.
        r (str): set video rate

            Defaults to 25.
        duration (str): set video duration

            Defaults to -0.000001.
        d (str): set video duration

            Defaults to -0.000001.
        sar (str): set video sample aspect ratio (from 0 to INT_MAX)

            Defaults to 1/1.

    Returns:
        "Stream": The output stream.
    """
    return create_source(
        filter_name="smptebars",
        named_arguments={
            "size": size,
            "s": s,
            "rate": rate,
            "r": r,
            "duration": duration,
            "d": d,
            "sar": sar,
        },
    )


def smptehdbars(
    size: str | None = None,
    s: str | None = None,
    rate: str | None = None,
    r: str | None = None,
    duration: str | None = None,
    d: str | None = None,
    sar: str | None = None,
) -> "Stream":
    """Generate SMPTE HD color bars.
    Args:
        size (str): set video size

            Defaults to 320x240.
        s (str): set video size

            Defaults to 320x240.
        rate (str): set video rate

            Defaults to 25.
        r (str): set video rate

            Defaults to 25.
        duration (str): set video duration

            Defaults to -0.000001.
        d (str): set video duration

            Defaults to -0.000001.
        sar (str): set video sample aspect ratio (from 0 to INT_MAX)

            Defaults to 1/1.

    Returns:
        "Stream": The output stream.
    """
    return create_source(
        filter_name="smptehdbars",
        named_arguments={
            "size": size,
            "s": s,
            "rate": rate,
            "r": r,
            "duration": duration,
            "d": d,
            "sar": sar,
        },
    )


def testsrc(
    size: str | None = None,
    s: str | None = None,
    rate: str | None = None,
    r: str | None = None,
    duration: str | None = None,
    d: str | None = None,
    sar: str | None = None,
    decimals: int | None = None,
    n: int | None = None,
) -> "Stream":
    """Generate test pattern.
    Args:
        size (str): set video size

            Defaults to 320x240.
        s (str): set video size

            Defaults to 320x240.
        rate (str): set video rate

            Defaults to 25.
        r (str): set video rate

            Defaults to 25.
        duration (str): set video duration

            Defaults to -0.000001.
        d (str): set video duration

            Defaults to -0.000001.
        sar (str): set video sample aspect ratio (from 0 to INT_MAX)

            Defaults to 1/1.
        decimals (int): set number of decimals to show (from 0 to 17)

            Defaults to 0.
        n (int): set number of decimals to show (from 0 to 17)

            Defaults to 0.

    Returns:
        "Stream": The output stream.
    """
    return create_source(
        filter_name="testsrc",
        named_arguments={
            "size": size,
            "s": s,
            "rate": rate,
            "r": r,
            "duration": duration,
            "d": d,
            "sar": sar,
            "decimals": decimals,
            "n": n,
        },
    )


def testsrc2(
    size: str | None = None,
    s: str | None = None,
    rate: str | None = None,
    r: str | None = None,
    duration: str | None = None,
    d: str | None = None,
    sar: str | None = None,
    alpha: int | None = None,
) -> "Stream":
    """Generate another test pattern.
    Args:
        size (str): set video size

            Defaults to 320x240.
        s (str): set video size

            Defaults to 320x240.
        rate (str): set video rate

            Defaults to 25.
        r (str): set video rate

            Defaults to 25.
        duration (str): set video duration

            Defaults to -0.000001.
        d (str): set video duration

            Defaults to -0.000001.
        sar (str): set video sample aspect ratio (from 0 to INT_MAX)

            Defaults to 1/1.
        alpha (int): set global alpha (opacity) (from 0 to 255)

            Defaults to 255.

    Returns:
        "Stream": The output stream.
    """
    return create_source(
        filter_name="testsrc2",
        named_arguments={
            "size": size,
            "s": s,
            "rate": rate,
            "r": r,
            "duration": duration,
            "d": d,
            "sar": sar,
            "alpha": alpha,
        },
    )


def yuvtestsrc(
    size: str | None = None,
    s: str | None = None,
    rate: str | None = None,
    r: str | None = None,
    duration: str | None = None,
    d: str | None = None,
    sar: str | None = None,
) -> "Stream":
    """Generate YUV test pattern.
    Args:
        size (str): set video size

            Defaults to 320x240.
        s (str): set video size

            Defaults to 320x240.
        rate (str): set video rate

            Defaults to 25.
        r (str): set video rate

            Defaults to 25.
        duration (str): set video duration

            Defaults to -0.000001.
        d (str): set video duration

            Defaults to -0.000001.
        sar (str): set video sample aspect ratio (from 0 to INT_MAX)

            Defaults to 1/1.

    Returns:
        "Stream": The output stream.
    """
    return create_source(
        filter_name="yuvtestsrc",
        named_arguments={
            "size": size,
            "s": s,
            "rate": rate,
            "r": r,
            "duration": duration,
            "d": d,
            "sar": sar,
        },
    )


def zoneplate(
    size: str | None = None,
    s: str | None = None,
    rate: str | None = None,
    r: str | None = None,
    duration: str | None = None,
    d: str | None = None,
    sar: str | None = None,
    precision: int | None = None,
    xo: int | None = None,
    yo: int | None = None,
    to: int | None = None,
    k0: int | None = None,
    kx: int | None = None,
    ky: int | None = None,
    kt: int | None = None,
    kxt: int | None = None,
    kyt: int | None = None,
    kxy: int | None = None,
    kx2: int | None = None,
    ky2: int | None = None,
    kt2: int | None = None,
    ku: int | None = None,
    kv: int | None = None,
) -> "Stream":
    """Generate zone-plate.
    Args:
        size (str): set video size

            Defaults to 320x240.
        s (str): set video size

            Defaults to 320x240.
        rate (str): set video rate

            Defaults to 25.
        r (str): set video rate

            Defaults to 25.
        duration (str): set video duration

            Defaults to -0.000001.
        d (str): set video duration

            Defaults to -0.000001.
        sar (str): set video sample aspect ratio (from 0 to INT_MAX)

            Defaults to 1/1.
        precision (int): set LUT precision (from 4 to 16)

            Defaults to 10.
        xo (int): set X-axis offset (from INT_MIN to INT_MAX)

            Defaults to 0.
        yo (int): set Y-axis offset (from INT_MIN to INT_MAX)

            Defaults to 0.
        to (int): set T-axis offset (from INT_MIN to INT_MAX)

            Defaults to 0.
        k0 (int): set 0-order phase (from INT_MIN to INT_MAX)

            Defaults to 0.
        kx (int): set 1-order X-axis phase (from INT_MIN to INT_MAX)

            Defaults to 0.
        ky (int): set 1-order Y-axis phase (from INT_MIN to INT_MAX)

            Defaults to 0.
        kt (int): set 1-order T-axis phase (from INT_MIN to INT_MAX)

            Defaults to 0.
        kxt (int): set X-axis*T-axis product phase (from INT_MIN to INT_MAX)

            Defaults to 0.
        kyt (int): set Y-axis*T-axis product phase (from INT_MIN to INT_MAX)

            Defaults to 0.
        kxy (int): set X-axis*Y-axis product phase (from INT_MIN to INT_MAX)

            Defaults to 0.
        kx2 (int): set 2-order X-axis phase (from INT_MIN to INT_MAX)

            Defaults to 0.
        ky2 (int): set 2-order Y-axis phase (from INT_MIN to INT_MAX)

            Defaults to 0.
        kt2 (int): set 2-order T-axis phase (from INT_MIN to INT_MAX)

            Defaults to 0.
        ku (int): set 0-order U-color phase (from INT_MIN to INT_MAX)

            Defaults to 0.
        kv (int): set 0-order V-color phase (from INT_MIN to INT_MAX)

            Defaults to 0.

    Returns:
        "Stream": The output stream.
    """
    return create_source(
        filter_name="zoneplate",
        named_arguments={
            "size": size,
            "s": s,
            "rate": rate,
            "r": r,
            "duration": duration,
            "d": d,
            "sar": sar,
            "precision": precision,
            "xo": xo,
            "yo": yo,
            "to": to,
            "k0": k0,
            "kx": kx,
            "ky": ky,
            "kt": kt,
            "kxt": kxt,
            "kyt": kyt,
            "kxy": kxy,
            "kx2": kx2,
            "ky2": ky2,
            "kt2": kt2,
            "ku": ku,
            "kv": kv,
        },
    )



import subprocess
import json
from typing import Any
from pyffmpeg.errors import Error
from pyffmpeg._utils import convert_kwargs_to_cmd_line_args


def probe(
    filename: str, cmd: str = "ffprobe", timeout: float | None = None, **kwargs
) -> dict[str, Any]:
    """
    Runs ffprobe and returns parsed json result
    """
    options = ["-show_format", "-show_streams", "-of", "json"]
    options.extend(convert_kwargs_to_cmd_line_args(kwargs))
    args = [cmd, *options, filename]
    result = subprocess.run(args, capture_output=True, timeout=timeout, check=False)

    if result.returncode != 0:
        raise Error(
            "ffprobe error (see stderr output for detail)", result.stdout, result.stderr
        )

    return json.loads(result.stdout)



from typing import Sequence, Any

from pyffmpeg.node import (
    FilterMultiOutput,
    FilterNode,
    InputNode,
    RunnableNode,
    OutputNode,
    Stream,
    MergedOutputNode,
    GraphSorter,
    CommandBuilder,
)


def input(
    filename: str,
    format: str | None = None,
    ss: str | float | None = None,
    t: str | float | None = None,
    to: str | float | None = None,
    r: str | int | float | None = None,
    video_size: str | tuple[int, int] | None = None,
    pix_fmt: str | None = None,
    hwaccel: str | None = None,
    loop: bool | int | None = None,
    re: bool = False,
    **kwargs,
) -> Stream:
    """Creates an input stream from a filename with optional input-specific flags.

    This function serves as the entry point for creating a processing graph.
    It accepts common FFmpeg input options as explicit arguments.
    Less common options can be passed via kwargs.

    Args:
        filename (str): Path to the input file, URL, or device.
        format (str | None): Force input format (ffmpeg flag: ``-f``).
            Example: ``'image2'``, ``'alsa'``, ``'v4l2'``, ``'rawvideo'``.
        ss (str | float | None): Seeks in this input file to position.
            Note: When used as an input option, this seeks BEFORE decoding
            (fast but less accurate).
        t (str | float | None): Limit the duration of data read from the input file.
        to (str | float | None): Stop reading the input after a specific time position.
        r (str | int | float | None): Set frame rate (Hz value, fraction or abbreviation).
            Crucial for raw video or image sequences.
        video_size (str | tuple[int, int] | None): Set frame size.
            Can be a string (e.g., ``'1920x1080'``) or a tuple ``(1920, 1080)``.
            Required for raw video.
        pix_fmt (str | None): Set input pixel format. Required for raw video (e.g., ``'yuv420p'``).
        hwaccel (str | None): Use hardware acceleration to decode the matching stream(s).
            Example: ``'cuda'``, ``'videotoolbox'``, ``'vaapi'``.
        loop (bool | int | None): Loop over the input stream. Useful for images.
            Set to ``True`` or ``1`` to enable looping.
        re (bool): Read input at native frame rate (ffmpeg flag: ``-re``).
            Mainly used for simulating a grab device or live input stream.
            Defaults to ``False``.
        **kwargs: Additional input options specific to the demuxer or decoder.
            Example: ``input("file.mp4", log_level="debug")``.

    Returns:
        Stream: Output stream of the created InputNode.
    """
    options: dict[str, Any] = kwargs.copy()

    if re:
        options["re"] = None

    if format is not None:
        options["format"] = format
    if ss is not None:
        options["ss"] = ss
    if t is not None:
        options["t"] = t
    if to is not None:
        options["to"] = to
    if r is not None:
        options["r"] = r
    if video_size is not None:
        options["video_size"] = video_size
    if pix_fmt is not None:
        options["pix_fmt"] = pix_fmt
    if hwaccel is not None:
        options["hwaccel"] = hwaccel
    if loop is not None:
        options["loop"] = loop

    return InputNode(filename, options).output_streams[0]


def merge_outputs(*nodes: "RunnableNode") -> "MergedOutputNode":
    """
    Groups multiple runnable nodes (outputs or sinks) into a single execution unit.

    Args:
        *nodes (RunnableNode): Variable list of graph endpoints. Accepts both standard OutputNode
            and SinkNode objects.

    Returns:
        MergedOutputNode: An aggregate object that can be run as a single process.
    """
    return MergedOutputNode(nodes)


def get_args(
    output: OutputNode | Sequence[OutputNode] | MergedOutputNode, **global_options
) -> list[str]:
    """Creates command arguments for the output or a list of outputs"""
    if isinstance(output, Sequence):
        # output is reversed to satisfy tests order requirements
        output = MergedOutputNode(outputs=tuple(reversed(output)))
    sorter = GraphSorter(output)
    command_builder = CommandBuilder(
        sorter.sort(),
        output.global_options + output._compile_global_kwargs(global_options),
    )
    return command_builder.build_args()


def filter(
    stream_or_streams_list: Stream | list[Stream], filter_name: str, *args, **kwargs
) -> "Stream":
    """Custom filter with single or many inputs and a single output"""
    if isinstance(stream_or_streams_list, (list, tuple)):
        if stream_or_streams_list and isinstance(stream_or_streams_list[0], Stream):
            return stream_or_streams_list[0]._apply_filter(
                filter_name, args, kwargs, stream_or_streams_list
            )
    if isinstance(stream_or_streams_list, Stream):
        return stream_or_streams_list.filter(filter_name, *args, **kwargs)


def filter_multi_output(
    stream_or_streams_list: Stream | list[Stream], filter_name: str, *args, **kwargs
) -> "FilterMultiOutput":
    """Creates a custom filter allowing dynamic creation of output streams, accepts many inputs"""
    if isinstance(stream_or_streams_list, (list, tuple)):
        inputs = stream_or_streams_list
    elif isinstance(stream_or_streams_list, Stream):
        inputs = [stream_or_streams_list]
    node = FilterNode(
        filter_name=filter_name,
        positional_arguments=args,
        named_arguments=kwargs,
        inputs=inputs,
        num_output_streams=0,
    )
    return FilterMultiOutput(node)



import subprocess
from pathlib import Path

from pyffmpeg.parser import Parser
from pyffmpeg.generator import CodeGenerator

FILTERS_OUTPUT = Path("src/pyffmpeg/generated_filters.py")
SOURCES_OUTPUT = Path("src/pyffmpeg/sources.py")


def get_filters_to_process() -> list[dict]:
    """
    Gets all filter names for processing from ffmpeg -filters.
    Identifies sources (|->) and standard filters (->).
    Includes sinks (->|).
    """
    try:
        result = subprocess.run(["ffmpeg", "-filters"], capture_output=True, text=True)
    except FileNotFoundError:
        print("❌ Did not find ffmpeg")
        return []

    filters_metadata = []
    seen = set()

    for line in result.stdout.splitlines():
        parts = line.split()

        if len(parts) < 3:
            continue
        name = parts[1]
        io_specifier = parts[2]

        if "->" not in io_specifier:
            continue

        if name in seen:
            continue
        seen.add(name)

        item = {"name": name}

        if io_specifier.startswith("|"):
            item["is_source"] = True
        else:
            item["is_source"] = False

        filters_metadata.append(item)

    filters_metadata.sort(key=lambda x: x["name"])
    return filters_metadata


def get_ffmpeg_help(filter_name: str) -> str:
    """Gets raw text from ffmpeg help."""
    try:
        result = subprocess.run(
            ["ffmpeg", "--help", f"filter={filter_name}"],
            capture_output=True,
            text=True,
            check=True,
        )
        return result.stdout
    except subprocess.CalledProcessError:
        print(f"⚠️  Couldn't get help for: {filter_name}")
        return ""


def generate_files():
    """Generates filters to generated_filters.py and sources to sources.py"""
    print(f"Starting generation...")
    print(f" - Filters -> {FILTERS_OUTPUT}")
    print(f" - Sources -> {SOURCES_OUTPUT}")

    with (
        open(FILTERS_OUTPUT, "w", encoding="utf-8") as f_filters,
        open(SOURCES_OUTPUT, "w", encoding="utf-8") as f_sources,
    ):
        f_filters.write("# --- AUTO-GENERATED FILE ---\n")
        f_filters.write("from typing import TYPE_CHECKING, Literal\n\n")
        f_filters.write("if TYPE_CHECKING:\n")
        f_filters.write(
            "    from pyffmpeg.node import Stream, FilterMultiOutput, SinkNode\n\n"
        )

        f_filters.write("class GeneratedFiltersMixin:\n")
        f_filters.write('    """\n')
        f_filters.write("    Mixin class containing auto-generated filter methods.\n")
        f_filters.write("    This class should be inherited by the Stream class.\n")
        f_filters.write('    """\n')

        f_sources.write("# --- AUTO-GENERATED FILE ---\n")
        f_sources.write("from typing import TYPE_CHECKING, Literal\n")
        f_sources.write("from pyffmpeg._utils import create_source\n\n")
        f_sources.write("if TYPE_CHECKING:\n")
        f_sources.write("    from pyffmpeg.node import Stream, FilterMultiOutput\n\n")

        items = get_filters_to_process()
        total_count = len(items)
        print(f"     Found {total_count} items to process.")

        success_count = 0
        for item in items:
            name = item["name"]
            is_source = item["is_source"]

            print(f"    Parsing: {name}...", end=" ")

            help_text = get_ffmpeg_help(name)
            if not help_text:
                print("❌ No help")
                continue

            try:
                parser = Parser(help_text)
                data = parser.parse()

                generator = CodeGenerator(data, is_source=is_source)
                code = generator.generate()

                if is_source:
                    f_sources.write(code + "\n\n")
                else:
                    f_filters.write(code + "\n\n")

                print("✅")
                success_count += 1

            except Exception as e:
                print(f"❌ Error: {e}")

    print(f"\n Finished. Generated {success_count}/{total_count} items.")


if __name__ == "__main__":
    generate_files()



import inspect
import functools
from typing import Any
from collections.abc import Callable, Sequence
from pyffmpeg._run import run, compile
from pyffmpeg import _run
from pyffmpeg._utils import escape_text_content
from pyffmpeg.io import input, merge_outputs, get_args
from pyffmpeg.probe import probe
from pyffmpeg.errors import Error
from pyffmpeg.node import (
    Node,
    OutputNode,
    Stream,
    TypedStream,
    IndexedStream,
    FilterNode,
    FilterMultiOutput,
)


def __getattr__(name: str) -> Callable[..., Any]:
    """
    Lets call methods of StreamCompatWrapper or Node as a standalone functions.
    First argument should be of type StreamCompatWrapper, Node or list list of StreamCompatWrapper.
    """

    def wrapper(
        obj: StreamCompatWrapper | Node | list[StreamCompatWrapper], *args, **kwargs
    ) -> Any:
        target_obj = obj
        if isinstance(obj, (list, tuple)) and obj:
            target_obj = obj[0]
            inputs_list = obj[1:]
        if not isinstance(target_obj, (StreamCompatWrapper, Node)):
            raise TypeError(
                f"Cannot call '{name}'. First argument "
                f"is '{type(obj)}', but must be StreamCompatWrapper, Node or a list of StreamCompatWrapper."
            )
        try:
            method = getattr(target_obj, name)
        except AttributeError:
            raise AttributeError(
                f"Module 'pyffmpeg._compat' does not have a function '{name}' "
                f"and an object of type '{type(target_obj)}' does not have a method with that name."
            )
        sig = inspect.signature(method)
        if "inputs" in sig.parameters:
            args = (inputs_list, *args)
            return method(*args, **kwargs)
        else:
            return method(*args, **kwargs)

    return wrapper


def wrap_stream_output(func):
    """Wraps output of func which is expected to be a stream or a sequence of streams in a StreamCompatWrapper"""

    @functools.wraps(func)
    def wrapper(*args, **kwargs):
        return_object = func(*args, **kwargs)
        if isinstance(return_object, Stream):
            return StreamCompatWrapper(return_object)
        if isinstance(return_object, Sequence) and all(
            isinstance(stream, (Stream | TypedStream | IndexedStream))
            for stream in return_object
        ):
            return [StreamCompatWrapper(stream) for stream in return_object]
        return return_object

    return wrapper


def wrap_stream_input(func):
    """Converts any argument of type StreamCompatWrapper to stream it contains and calls func"""

    @functools.wraps(func)
    def wrapper(*args, **kwargs):
        args = [
            arg.stream if isinstance(arg, StreamCompatWrapper) else arg for arg in args
        ]
        return func(*args, **kwargs)

    return wrapper


input = wrap_stream_output(input)


class StreamCompatWrapper:
    """Encapsulates the stream to invoke its methods underneath while providing backward compatible API."""

    def __init__(self, stream: Stream | TypedStream | IndexedStream):
        self.stream: Stream | TypedStream | IndexedStream = stream

    def __getattr__(self, name):
        """Returns wrapped stream method or wrapped stream attribute."""
        if hasattr(self.stream, name):
            attr = getattr(self.stream, name)

            if callable(attr):
                return wrap_stream_input(wrap_stream_output(attr))
            if isinstance(attr, (TypedStream, IndexedStream)):
                return StreamCompatWrapper(attr)
            if isinstance(attr, Sequence) and all(isinstance(x, Stream) for x in attr):
                return [StreamCompatWrapper(stream) for stream in attr]

    def __eq__(self, other: object) -> bool:
        if not isinstance(other, StreamCompatWrapper):
            return NotImplemented
        return self.stream == other.stream

    def __hash__(self) -> int:
        return hash(self.stream)

    def __getitem__(self, key: str) -> "StreamCompatWrapper":
        return StreamCompatWrapper(self.stream[key])

    @wrap_stream_input
    def output(self, *args, **kwargs) -> "OutputNode":
        streams = []
        filename = None
        streams.append(self)

        for arg in args:
            if isinstance(arg, Stream):
                streams.append(arg)
            elif isinstance(arg, str):
                filename = arg
            else:
                raise TypeError(f"Unexpected argument type: {type(arg)}")

        if not filename:
            raise ValueError("No output filename provided to output()")

        return OutputNode(filename, streams, output_options=kwargs)

    def filter(self, filter_name: str, *args, **kwargs) -> "StreamCompatWrapper":
        """Custom filter with a single input and a single output"""
        return StreamCompatWrapper(
            self.stream._apply_filter(filter_name, args, kwargs)[0]
        )

    @wrap_stream_output
    @wrap_stream_input
    def scale(self, height: int, width: int) -> "Stream":
        """Scales to width and height."""
        return self._apply_filter(
            "scale", named_arguments={"height": str(height), "width": str(width)}
        )[0]

    @wrap_stream_output
    @wrap_stream_input
    def split(self, num_outputs: int = 2) -> list["Stream"]:
        """Split into multiple identical streams."""
        return self._apply_filter(
            "split", postional_arguments=(num_outputs,), num_output_streams=num_outputs
        )

    @wrap_stream_output
    @wrap_stream_input
    def asplit(self, num_outputs: int = 2) -> list["Stream"]:
        """Split into multiple identical audio streams."""
        return self._apply_filter(
            "asplit", postional_arguments=(num_outputs,), num_output_streams=num_outputs
        )

    @wrap_stream_output
    @wrap_stream_input
    def overlay(
        self,
        overlay_stream: "Stream",
        x: int | None = None,
        y: int | None = None,
        eof_action: str = "repeat",
    ) -> "Stream":
        """Overlay another video stream on top of this one."""
        named_arguments = {"eof_action": eof_action}
        if x:
            named_arguments["x"] = str(x)
        if y:
            named_arguments["y"] = str(y)

        return self._apply_filter(
            "overlay",
            named_arguments=named_arguments,
            inputs=[self, overlay_stream],
        )[0]

    @wrap_stream_output
    @wrap_stream_input
    def trim(self, **kwargs) -> "Stream":
        return self._apply_filter("trim", named_arguments=kwargs)[0]

    @wrap_stream_output
    @wrap_stream_input
    def vflip(self) -> "Stream":
        """Flip the input video vertically"""
        return self._apply_filter(filter_name="vflip")[0]

    @wrap_stream_output
    @wrap_stream_input
    def hflip(self) -> "Stream":
        """Flip the input video horizontally"""
        return self._apply_filter(filter_name="hflip")[0]

    @wrap_stream_output
    @wrap_stream_input
    def crop(self, x: int, y: int, width: int, height: int) -> "Stream":
        """Crop the input video to given dimensions"""
        return self._apply_filter("crop", postional_arguments=(width, height, x, y))[0]

    @wrap_stream_output
    @wrap_stream_input
    def concat(self, *streams: "Stream", **kwargs) -> "Stream":
        """Concatenate audio and video streams, joining them together one after the other"""
        video_stream_count = kwargs.get("v", 1)
        audio_stream_count = kwargs.get("a", 0)
        output_stream_count = video_stream_count + audio_stream_count
        input_stream_count = int(len([self, *streams]))
        if (input_stream_count % output_stream_count) != 0:
            raise ValueError(
                f"Expected concat input streams to have length multiple of {output_stream_count} (v={video_stream_count}, a={audio_stream_count}); got {input_stream_count}"
            )
        kwargs["n"] = int(input_stream_count / output_stream_count)
        return self._apply_filter(
            "concat",
            named_arguments=kwargs,
            inputs=[self, *streams],
            num_output_streams=output_stream_count,
        )[0]

    @wrap_stream_output
    @wrap_stream_input
    def drawbox(
        self,
        x: int,
        y: int,
        width: int,
        height: int,
        color: str,
        thickness: int | None = None,
    ) -> "Stream":
        """Draw a colored box on the input image"""
        return self._apply_filter(
            "drawbox",
            postional_arguments=(x, y, width, height, color),
            named_arguments={"t": str(thickness)},
        )[0]

    @wrap_stream_output
    @wrap_stream_input
    def drawtext(self, text: str, **kwargs) -> "Stream":
        kwargs["text"] = escape_text_content(text)
        return self._apply_filter("drawtext", named_arguments=kwargs)[0]


def filter(
    stream_spec: list[StreamCompatWrapper] | StreamCompatWrapper,
    filter_name: str,
    *args,
    **kwargs,
) -> StreamCompatWrapper:
    if isinstance(stream_spec, StreamCompatWrapper):
        stream_spec = [stream_spec.stream]
    if all(isinstance(s, StreamCompatWrapper) for s in stream_spec):
        stream_spec = [s.stream for s in stream_spec]
    return StreamCompatWrapper(
        stream_spec[0]._apply_filter(filter_name, args, kwargs, inputs=stream_spec)[0]
    )


def filter_multi_output(
    stream_spec: list[StreamCompatWrapper] | StreamCompatWrapper,
    filter_name: str,
    *args,
    **kwargs,
) -> StreamCompatWrapper:
    if isinstance(stream_spec, StreamCompatWrapper):
        stream_spec = [stream_spec.stream]
    if all(isinstance(s, StreamCompatWrapper) for s in stream_spec):
        stream_spec = [s.stream for s in stream_spec]
    node = FilterNode(
        filter_name=filter_name,
        positional_arguments=args,
        named_arguments=kwargs,
        inputs=stream_spec,
        num_output_streams=0,
    )
    return FilterMultiOutput(node)



from pyffmpeg.node import RunnableNode, OutputNode, MergedOutputNode


def compile(
    node: RunnableNode,
    cmd: str = "ffmpeg",
    overwrite_output: bool = False,
    **global_options,
) -> list[str]:
    return node.compile(cmd=cmd, overwrite_output=overwrite_output, **global_options)


def run(
    node_or_node_list: RunnableNode | list[OutputNode],
    cmd: str | list[str] = "ffmpeg",
    capture_stdout: bool = False,
    capture_stderr: bool = False,
    input: bytes | None = None,
    quiet: bool = False,
    overwrite_output: bool = False,
    cwd: str | None = None,
) -> tuple[bytes | None, bytes | None]:
    """Executes ffmpeg command."""
    if isinstance(node_or_node_list, (list, tuple)) and all(
        [isinstance(n, OutputNode) for n in node_or_node_list]
    ):
        node_or_node_list = MergedOutputNode(node_or_node_list)
    return node_or_node_list.run(
        cmd=cmd,
        capture_stdout=capture_stdout,
        capture_stderr=capture_stderr,
        input=input,
        quiet=quiet,
        overwrite_output=overwrite_output,
        cwd=cwd,
        compile_function=compile,
    )



from pprint import pprint
import subprocess
from typing import Any
import pytest
from pyffmpeg.parser import Parser


def get_parsed_filter_data(filter_name: str) -> dict[str, Any]:
    """Runs ffmpeg help for a given filter, parses output and returns data"""
    cmd = ["ffmpeg", "--help", f"filter={filter_name}"]

    result = subprocess.run(cmd, capture_output=True, text=True)
    assert result.returncode == 0

    parser = Parser(result.stdout)
    return parser.parse()


def test_overlay_filter_name():
    data = get_parsed_filter_data("overlay")
    assert data.get("filter_name") == "overlay"


def test_scale_filter_name():
    data = get_parsed_filter_data("scale")
    assert data.get("filter_name") == "scale"


def test_overlay_description():
    data = get_parsed_filter_data("overlay")
    assert data.get("description") == "Overlay a video source on top of the input."


def test_scale_description():
    data = get_parsed_filter_data("scale")
    assert (
        data.get("description")
        == "Scale the input video size and/or convert the image format."
    )


def test_overlay_inputs():
    data = get_parsed_filter_data("overlay")
    assert data.get("inputs") == [
        {"name": "main_stream", "type": "video"},
        {"name": "overlay_stream", "type": "video"},
    ]


def test_scale_inputs():
    data = get_parsed_filter_data("scale")
    assert data.get("inputs") == [{"name": "default_stream", "type": "video"}]


def test_overlay_outputs():
    data = get_parsed_filter_data("overlay")
    assert data.get("outputs") == [{"name": "default_stream", "type": "video"}]


def test_scale_outputs():
    data = get_parsed_filter_data("scale")
    assert data.get("outputs") == [{"name": "default_stream", "type": "video"}]


def test_split_outputs():
    data = get_parsed_filter_data("split")
    print(data)
    assert data.get("outputs") == []


def test_overlay_option_block():
    data = get_parsed_filter_data("overlay")
    assert data["options"][0] == {
        "name": "x",
        "type": "string",
        "description": "set the x expression",
        "default": "0",
        "choices": [],
    }
    assert data["options"][2] == {
        "name": "eof_action",
        "type": "int",
        "description": "Action to take when encountering EOF from secondary input  (from 0 to 2)",
        "default": "repeat",
        "choices": [
            {
                "name": "repeat",
                "value": "0",
                "description": "Repeat the previous frame.",
            },
            {"name": "endall", "value": "1", "description": "End both streams."},
            {
                "name": "pass",
                "value": "2",
                "description": "Pass through the main input.",
            },
        ],
    }


def test_split_all_data():
    """an example of dynamic outputs filter"""
    data = get_parsed_filter_data("split")
    assert data == {
        "description": "Pass on the input to N video outputs.",
        "filter_name": "split",
        "inputs": [{"name": "default_stream", "type": "video"}],
        "outputs": [],
        "is_dynamic_inputs": False,
        "is_dynamic_outputs": True,
        "options": [
            {
                "choices": [],
                "default": "2",
                "description": "set number of outputs (from 1 to INT_MAX)",
                "name": "outputs",
                "type": "int",
            }
        ],
    }


def test_concat():
    """an example of dynamic inputs filter (and dynamic outputs)"""
    data = get_parsed_filter_data("concat")
    assert data["inputs"] == []
    assert data["is_dynamic_inputs"] == True

    assert data["outputs"] == []
    assert data["is_dynamic_outputs"] == True


def test_sierpinski():
    """an example of source filter"""
    data = get_parsed_filter_data("sierpinski")
    assert data["inputs"] == []
    assert data["is_dynamic_inputs"] == False
    assert len(data["outputs"]) == 1


def test_nullsink():
    """an example of sink filter"""
    data = get_parsed_filter_data("nullsink")
    assert data["outputs"] == []
    assert data["is_dynamic_outputs"] == False
    assert len(data["inputs"]) == 1
